// WL_AGENT.C

#include "wl_def.h"
#pragma hdrstop

void DrawScaleds (void);

extern   statetype s_rocket;

/*
=============================================================================

                                LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

                                GLOBAL VARIABLES

=============================================================================
*/
// player state info
int32_t         thrustspeed;
word            plux,pluy;          // player coordinates scaled to unsigned
short           anglefrac;
objtype        *LastAttacker;

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/
void    T_Player (objtype *ob);
void    T_Attack (objtype *ob);
void    T_Teleport (int x,int y);

void    DrawPlayerWeapon (void);
void InitAreas (void);
void ConnectAreas (void);
void RecursiveConnect (int areanumber);
void A_DeathScream (objtype *ob);

statetype   s_player = {false,0,0,(statefunc) T_Player,NULL,NULL};
statetype   s_attack = {false,0,0,(statefunc) T_Attack,NULL,NULL};

struct atkinf { int8_t    tics,attack,frame; } 
attackinfo[NUMWEAPONS][12] = // attack is 1 for gun, 2 for knife
{
    { {6,0,1},{6,2,2},{6,0,3},{6,-1,4} }, // Knife
#ifdef NEWGUNCODE
    { {6,99,1},{6,1,2},{6,0,3},{6,-1,4} }, // Pistol
    { {6,99,1},{6,1,2},{6,3,3},{6,-1,4} }, // Machinegun
    { {6,99,1},{6,1,2},{6,4,3},{6,-1,4} }, // Chaingun
    { {6,99,1},{6,1,2},{6,0,3},{6,-1,4} }, // Rocket launcher
#else
    { {6,0,1},{6,1,2},{6,0,3},{6,-1,4} }, // Pistol
    { {6,0,1},{6,1,2},{6,3,3},{6,-1,4} }, // Machinegun
    { {6,0,1},{6,1,2},{6,4,3},{6,-1,4} }, // Chaingun
    { {6,0,1},{6,1,2},{6,0,3},{6,-1,4} }, // Rocket launcher
#endif
};

//===========================================================================

//----------

void Attack (void);
void Use (void);
void Search (objtype *ob);
void SelectWeapon (void);
void SelectItem (void);

//----------

boolean TryMove (objtype *ob);
void T_Player (objtype *ob);

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove);

#ifdef KILLSDISPLAY
void DrawKills (void);
#endif

/*
=============================================================================

                                CONTROL STUFF

=============================================================================
*/

/*
======================
=
= CheckWeaponChange
=
= Keys 1-4 change weapons
=
======================
*/
void CheckWeaponChange (void)
{
    int newWeapon = -1;

    if (ffDataBottomRight!=0 && !gamestate.ammo) return;
#ifdef _arch_dreamcast
    int joyx, joyy;
    IN_GetJoyFineDelta (&joyx, &joyy);
    if(joyx < -64)
        buttonstate[bt_prevweapon] = true;
    else if(joyx > 64)
        buttonstate[bt_nextweapon] = true;
#endif

    if(buttonstate[bt_nextweapon] && !buttonheld[bt_nextweapon])
    {
        newWeapon = gamestate.weapon + 1;
        if(newWeapon > gamestate.bestweapon) newWeapon = 0;
    }
    else if(buttonstate[bt_prevweapon] && !buttonheld[bt_prevweapon])
    {
        newWeapon = gamestate.weapon - 1;
        if(newWeapon < 0) newWeapon = gamestate.bestweapon;
    }
    else
    {
        for(int i = wp_knife; i <= gamestate.bestweapon; i++)
        {
            if (buttonstate[bt_readyknife + i - wp_knife])
            {
                newWeapon = i;
                break;
            }
        }
    }

    if(newWeapon != -1)
    {
        gamestate.weapon = gamestate.chosenweapon = (weapontype) newWeapon;
        DrawWeapon();
    }
}

/* ============== BJ Direction ============== */
char BJ_Facing (void)
{
  char dir;
  if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8) { dir = di_east; }
  else if (player->angle < 3*ANGLES/8) { dir = di_north; }
  else if (player->angle < 5*ANGLES/8) { dir = di_west; }
  else { dir = di_south; }
  return dir;
}

/*
=======================
=
= ControlMovement
=
= Takes controlx,controly, and buttonstate[bt_strafe]
=
= Changes the player's angle and position
=
= There is an angle hack because when going 70 fps, the roundoff becomes
= significant
=
=======================
*/
void ControlMovement (objtype *ob)
{
    int32_t oldx,oldy;
    int     angle;
    int     angleunits;

if (demoplayback || demorecord) { gamestate.bjstrafe = 0;  gamestate.bjthrust =0; }

    thrustspeed = 0;

    oldx = player->x;
    oldy = player->y;

    if(buttonstate[bt_strafeleft])
    {
        angle = ob->angle + ANGLES/4;
        if(angle >= ANGLES)
            angle -= ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics);
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    if(buttonstate[bt_straferight])
    {
        angle = ob->angle - ANGLES/4;
        if(angle < 0)
            angle += ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics );
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    //
    // side to side move
    //
    if (buttonstate[bt_strafe] or gamestate.bjstrafe)
    {
        //
        // strafing
        //
        //
        if (controlx > 0)
        {
            angle = ob->angle - ANGLES/4;
            if (angle < 0)
                angle += ANGLES;
            Thrust (angle,controlx*MOVESCALE);      // move to left
        }
        else if (controlx < 0)
        {
            angle = ob->angle + ANGLES/4;
            if (angle >= ANGLES)
                angle -= ANGLES;
            Thrust (angle,-controlx*MOVESCALE);     // move to right
        }
    }
    else
    {
        //
        // not strafing
        //
        anglefrac += controlx;
        angleunits = anglefrac/ANGLESCALE;
        anglefrac -= angleunits*ANGLESCALE;
        ob->angle -= angleunits;

        if (ob->angle >= ANGLES)
            ob->angle -= ANGLES;
        if (ob->angle < 0)
            ob->angle += ANGLES;

    }

    //
    // forward/backwards move
    //
    if (controly < 0)
    {
        Thrust (ob->angle,-controly*MOVESCALE); // move forwards
    }
    else if (controly > 0)
    {
        angle = ob->angle + ANGLES/2;
        if (angle >= ANGLES)
            angle -= ANGLES;
        Thrust (angle,controly*BACKMOVESCALE);          // move backwards
    }

    if (gamestate.victoryflag)              // watching the BJ actor
        return;
}

/* ==================================================================
                         STATUS WINDOW STUFF
   ================================================================== */

/*
==================
=
= StatusDrawPic
=
==================
*/
void StatusDrawPic (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord ((screenWidth-scaleFactor*320)/16 + scaleFactor*x,
        screenHeight-scaleFactor*(STATUSLINES-y),picnum);
}

void StatusDrawFace(unsigned picnum)
{
    StatusDrawPic(FACEPICX, FACEPICY, picnum);
    StatusDrawLCD(picnum);
}

/*
==================
=
= DrawFace
=
==================
*/
void DrawFace (void)
{
    if(viewsize == 21 && ingame) return;
    if (SD_SoundPlaying() == GETGATLINGSND)
        StatusDrawFace(GOTGATLINGPIC);
    else if (gamestate.health)
    {
#ifdef SPEAR
        if (godmode)
            StatusDrawFace(GODMODEFACE1PIC+gamestate.faceframe);
        else
#endif
            StatusDrawFace(FACE1APIC+3*((MAXHEALTH-gamestate.health)/16)+gamestate.faceframe);
    }
    else
    {
#ifndef SPEAR
        if (LastAttacker && LastAttacker->obclass == needleobj)
            StatusDrawFace(MUTANTBJPIC);
        else
#endif
            StatusDrawFace(FACE8APIC);
    }
}

/*
===============
=
= UpdateFace
=
= Calls draw face if time to change
=
===============
*/
int facecount = 0;
int facetimes = 0;
int AmbientMem;
int ambientcount = 0;

void UpdateFace (void)
{
int AmbientX,AmbientY,RainLimit;

		if (!ffDataBottomRight) ambientcount += tics;
		if (ambientcount > 10l*70)
           {
           AmbientX = ((player->tilex)-1)+US_RndT()%3;
           AmbientY = ((player->tiley)-1)+US_RndT()%3;
#ifdef RNDFIX
#ifdef WOLFDXDEMO
           AmbientMem += (1+(US_RndT()%3));
           if (AmbientMem > 3) AmbientMem -= 4;
#else
           AmbientMem += (1+(US_RndT()%8));
           if (AmbientMem > 8) AmbientMem -= 9;
#endif
#endif
           ambientcount=0;
           if (!SD_SoundPlaying())
        {
#ifdef WOLFDXDEMO
            int soundsin[4]={ AMBIENTINDOOR1SND,
                AMBIENTINDOOR3SND,
                AMBIENTINDOOR6SND,
                AMBIENTINDOOR8SND
#else
            int soundsin[9]={ AMBIENTINDOOR0SND,
                AMBIENTINDOOR1SND,
                AMBIENTINDOOR2SND,
                AMBIENTINDOOR3SND,
                AMBIENTINDOOR4SND,
                AMBIENTINDOOR5SND,
                AMBIENTINDOOR6SND,
                AMBIENTINDOOR7SND,
                AMBIENTINDOOR8SND
#endif
                };
#ifdef WOLFDXDEMO
            int soundsout[4]={ AMBIENTOUTDOOR1SND,
                AMBIENTOUTDOOR4SND,
                AMBIENTOUTDOOR5SND,
                AMBIENTOUTDOOR7SND
#else
            int soundsout[9]={ AMBIENTOUTDOOR0SND,
                AMBIENTOUTDOOR1SND,
                AMBIENTOUTDOOR2SND,
                AMBIENTOUTDOOR3SND,
                AMBIENTOUTDOOR4SND,
                AMBIENTOUTDOOR5SND,
                AMBIENTOUTDOOR6SND,
                AMBIENTOUTDOOR7SND,
                AMBIENTOUTDOOR8SND
#endif
                };

            if (*(mapsegs[2]+((player->tiley<<mapshift))+(player->tilex))>>8) {
#ifdef RNDFIX
#ifdef WOLFDXDEMO
           if (gamestate.mapon < 1)
#endif
           PlaySoundLocGlobal ((soundsin[AmbientMem]),AmbientX,AmbientY);
#else
#ifdef WOLFDXDEMO
           if (gamestate.mapon < 1) PlaySoundLocGlobal ((soundsin[US_RndT()%4]),AmbientX,AmbientY);
#else
           PlaySoundLocGlobal ((soundsin[US_RndT()%9]),AmbientX,AmbientY);
#endif
#endif
           }
           else
#ifdef RNDFIX
           PlaySoundLocGlobal ((soundsout[AmbientMem]),AmbientX,AmbientY);
#else
#ifdef WOLFDXDEMO
           PlaySoundLocGlobal ((soundsout[US_RndT()%4]),AmbientX,AmbientY);
#else
           PlaySoundLocGlobal ((soundsout[US_RndT()%9]),AmbientX,AmbientY);
#endif
#endif
            }
}
    if (SD_SoundPlaying() == GETGATLINGSND)
        return;

#ifdef KILLSDISPLAY
    DrawKills(); 
#endif
  // don't make demo depend on sound playback
    if(demoplayback || demorecord)
    {
        if(facetimes > 0)
        {
            facetimes--;
            return;
        }
    }
    else if(SD_SoundPlaying() == GETGATLINGSND) return;

    facecount += tics;
    if (facecount > US_RndT())
    {
        gamestate.faceframe = (US_RndT()>>6);
        if (gamestate.faceframe==3) gamestate.faceframe = 1;

        facecount = 0;
        DrawFace ();
    }
}


/*
===============
=
= LatchNumber
=
= right justifies and pads with blanks
=
===============
*/
static void LatchNumber (int x, int y, unsigned width, int32_t number)
{
  unsigned length,c;
  char    str[20];

  ltoa (number,str,10);
  length = (unsigned) strlen (str);
  while (length<width)
  {
    StatusDrawPic (x,y,N_BLANKPIC);
    x++;
    width--;
  }
  c = length <= width ? 0 : length-width;
  while (c<length)
  {
    StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
    x++;
    c++;
  }
}


/*
===============
=
= DrawHealth
=
===============
*/
void DrawHealth (void)
{
  if(viewsize == 21 && ingame) return;
  LatchNumber (HEALTHX,HEALTHY,HEALTHD,gamestate.health);
}


/*
===============
=
= TakeDamage
=
===============
*/
void TakeDamage (int points,objtype *attacker)
{
  LastAttacker = attacker;
  if (gamestate.victoryflag) return;
  if (gamestate.difficulty==gd_baby) points>>=2;

  if (!godmode) gamestate.health -= points;

  if (gamestate.health<=0)
  {
    gamestate.health = 0;
    playstate = ex_died;
    killerobj = attacker;
  }

  if (godmode != 2) StartDamageFlash (points);

  DrawHealth ();
  DrawFace ();

#ifdef SPEAR  // MAKE BJ'S EYES BUG IF MAJOR DAMAGE!
  if (points > SPEARDAMAGEBUGEYE && 
      gamestate.health!=0 && !godmode && 
      viewsize != 21)
  {
    StatusDrawFace(BJOUCHPIC);
    facecount = 0;
  }
#endif
#ifdef TEKZOOMSHOCK     // BEGIN ZOOM BLOCK 
  StartDamageShock( points ); 
#endif                  // END ZOOM BLOCK 
}

/*
===============
=
= HealSelf
=
===============
*/
void HealSelf (int points)
{
  gamestate.health += points;
  if (gamestate.health>= MAXHEALTH) gamestate.health = MAXHEALTH;
  DrawHealth ();
  DrawFace ();
}

//===========================================================================

/*
===============
=
= DrawLevel
=
===============
*/
void DrawLevel (void)
{
  if(viewsize == 21 && ingame) return;
#ifdef SPEAR
  if (gamestate.mapon == SPEARLASTLEVEL)
    LatchNumber (LEVELX,LEVELY,LEVELD,SPEARLASTSHOWLEVEL);
  else
#endif
    LatchNumber (LEVELX,LEVELY,LEVELD,gamestate.mapon+1);
}

#ifdef KILLSDISPLAY
/* 
============ 
= 
= DrawKills 
= 
============ 
*/ 

void DrawKills (void) 
{ 
  if(viewsize == 21 && ingame) return;
  LatchNumber(KILLDISPLAYX,KILLDISPLAYY,KILLDISPLAYD,gamestate.killcount); 
}
#endif

//===========================================================================

/*
===============
=
= DrawLives
=
===============
*/
void DrawLives (void)
{
  if(viewsize == 21 && ingame) return;
  LatchNumber (LIVESX,LIVESY,LIVESD,gamestate.lives);
}

/*
===============
=
= GiveExtraMan
=
===============
*/
void GiveExtraMan (void)
{
  if (gamestate.lives<MAXLIVES) gamestate.lives++;
  DrawLives ();
  SD_PlaySound (BONUS1UPSND);
}

//===========================================================================

/*
===============
=
= DrawScore
=
===============
*/
void DrawScore (void)
{
  if(viewsize == 21 && ingame) return;
  LatchNumber (SCOREX,SCOREY,SCORED,gamestate.score);
}

/*
===============
=
= GivePoints
=
===============
*/
void GivePoints (int32_t points)
{
  gamestate.score += points;
  while (gamestate.score >= gamestate.nextextra)
  {
    gamestate.nextextra += EXTRAPOINTS;
    GiveExtraMan ();
  }
  DrawScore ();
}

//===========================================================================
#ifndef CHRISAUTOMAPTEK
/*
==================
=
= DrawWeapon
=
==================
*/
void DrawWeapon (void)
{
  if(viewsize == 21 && ingame) return;
  StatusDrawPic (WEAPONX,WEAPONY,KNIFEPIC+gamestate.weapon);
  DrawAmmo ();
}
#endif

/*
==================
=
= DrawKeys
=
==================
*/
void DrawKeys (void)
{
    if(viewsize == 21 && ingame) return;
    if (gamestate.keys & 1)
        StatusDrawPic (KEY1X,KEY1Y,KEY1PIC);
    else
        StatusDrawPic (KEY1X,KEY1Y,NOKEYPIC);

    if (gamestate.keys & 2)
        StatusDrawPic (KEY2X,KEY2Y,KEY2PIC);
    else
        StatusDrawPic (KEY2X,KEY2Y,NOKEYPIC);
#ifdef KEYS4
    if (gamestate.keys & 4)
        StatusDrawPic (KEY3X,KEY3Y,KEY3PIC);
    else
        StatusDrawPic (KEY3X,KEY3Y,NOKEYPIC);

    if (gamestate.keys & 8)
        StatusDrawPic (KEY4X,KEY4Y,KEY4PIC);
    else
        StatusDrawPic (KEY4X,KEY4Y,NOKEYPIC);
#endif
}

/*
==================
=
= GiveWeapon
=
==================
*/
void GiveWeapon (int weapon)
{
    if (gamestate.bestweapon<weapon) gamestate.bestweapon = 
                                     gamestate.weapon = 
                                     gamestate.chosenweapon = 
                                     (weapontype) weapon;
    DrawWeapon ();
}

//===========================================================================

/*
===============
=
= DrawAmmo
=
===============
*/

void DrawAmmo (void)
{
    if(viewsize == 21 && ingame) return;
if (ffDataBottomRight!=0)
   {
//    LatchNumber (27,16,2,gamestate.ammo);
    LatchNumber (26,16,3,gamestate.ammo);
    return;
    }
   switch (gamestate.weapon)
   {
      case wp_knife:
         StatusDrawPic(26, 16, N_NONUMPIC);
//         LatchNumber (26,16,3,0);
         break;
      case wp_pistol:
      case wp_machinegun:
         LatchNumber (26,16,3,gamestate.ammo);
         break;
      case wp_chaingun:
         LatchNumber (26,16,3,gamestate.ammo2);
         break;
#ifdef RLAUNCHERSTUFF
      case wp_rocketgun:
         LatchNumber (26,16,3,gamestate.rockets);
         break;
#endif
   }
}

/*
===============
=
= GiveAmmo
=
===============
*/
void GiveAmmo (int ammo)
{
    if (!gamestate.ammo)                            // knife was out
    {
        if (!gamestate.attackframe)
        {
            gamestate.weapon = gamestate.chosenweapon;
            DrawWeapon ();
        }
    }
    gamestate.ammo += ammo;
    if (gamestate.ammo > AMMOMAX) gamestate.ammo = AMMOMAX;
    DrawAmmo ();
}

//===========================================================================

/*
===============
=
= GiveRockets
=
===============
*/

void GiveRockets (int rockets)
{
    if (!gamestate.rockets)                            // knife was out
    {
        if (!gamestate.attackframe)
        {
            gamestate.weapon = gamestate.chosenweapon;
            DrawWeapon ();
        }
    }
    gamestate.rockets += rockets;
    if (gamestate.rockets > 8)
        gamestate.rockets = 8;
    DrawAmmo ();
}

/*
===============
=
= GiveAmmo2
=
===============
*/

void GiveAmmo2 (int ammo2)
{
    if (!gamestate.ammo2)                            // knife was out
    {
        if (!gamestate.attackframe)
        {
            gamestate.weapon = gamestate.chosenweapon;
            DrawWeapon ();
        }
    }
    gamestate.ammo2 += ammo2;
    if (gamestate.ammo2 > 200)
        gamestate.ammo2 = 200;
    DrawAmmo ();
}

/*
==================
=
= GiveKey
=
==================
*/
void GiveKey (int key)
{
    gamestate.keys |= (1<<key);
    DrawKeys ();
}


#ifdef STATBARCLOCK
/*
==================
   DrawTime
==================
*/
void DrawTime (void)
{        
	 int  sec, min;

    sec=gamestate.TimeCount/70;

    if (sec > 99*60) { sec = 99*60; } // 99 minutes max

	 min=sec/60;
	 sec%=60;

	 LatchNumber (25,4,1,(min/10));
	 LatchNumber (26,4,1,(min%10));
	 LatchNumber (28,4,1,(sec/10));
	 LatchNumber (29,4,1,(sec%10));
}
#endif
/*
=============================================================================

                                MOVEMENT

=============================================================================
*/

/*
===================
=
= GetBonus
=
===================
*/
void GetBonus (statobj_t *check)
{
    int     lineNumber;
    int     artnum;

    char    *text;

    switch (check->itemnumber)
    {
        case    bo_firstaid:
            if (gamestate.health >= MAXHEALTH) return;
            SD_PlaySound (HEALTH2SND);
            HealSelf (HEALTHAID);
            GetMessage("25 health restored by medkit.");
            break;

        case    bo_detect:
            if (gamestate.bombdetector) { GetMessage("You already have one."); return; }
            SD_PlaySound (PERCENT100SND);
            gamestate.bombdetector=true;
            GetMessage("An Ammo Crate booby-trap detector!  Yes!");
            break;

        case    bo_portmed:
            if (gamestate.medical > 5) { GetMessage("5 portable medkits is the maximum."); return; }

            SD_PlaySound (PORTSND);
            gamestate.medical++;
            GetMessage("Found a portable medkit!");
            break;

        case    bo_key1:
        case    bo_key2:
        case    bo_key3:
        case    bo_key4:
            GiveKey (check->itemnumber - bo_key1);
            SD_PlaySound (GETKEYSND);
            GetMessage("Key to locked door obtained.");
            break;

        case    bo_cross:
            SD_PlaySound (BONUS1SND);
            GivePoints (BONUSPOINT1);
            gamestate.treasurecount++;
            GetMessage("Cross: 100 points.");
            break;

        case    bo_chalice:
            SD_PlaySound (BONUS2SND);
            GivePoints (BONUSPOINT2);
            gamestate.treasurecount++;
            GetMessage("Chalice: 500 points.");
            break;

        case    bo_bible:
            SD_PlaySound (BONUS3SND);
            GivePoints (BONUSPOINT3);
            gamestate.treasurecount++;
            GetMessage("Chest of jewels: 1000 points.");
            break;

        case    bo_crown:
            SD_PlaySound (BONUS4SND);
            GivePoints (BONUSPOINT4);
            gamestate.treasurecount++;
            GetMessage("Crown: 5000 points!");
            break;

        case    bo_treas1:
                SD_PlaySound (BONUS1SND);
                GivePoints (400);
                gamestate.treasurecount++;
                GetMessage("Nazi pendant: 400 points.");
                break;
        case    bo_treas2:
                SD_PlaySound (TREASURE2SND);
                GivePoints (800);
                gamestate.treasurecount++;
                GetMessage("Red ruby: 800 points.");
                break;
        case    bo_treas3:
                SD_PlaySound (TREASURE3SND);
                GivePoints (1200);
                gamestate.treasurecount++;
                GetMessage("Gold bar: 1200 points.");
                break;
        case    bo_treas4:
                SD_PlaySound (TREASURE4SND);
                GivePoints (1600);
                gamestate.treasurecount++;
                GetMessage("Eagle figurine: 1600 points.");
                break;
        case    bo_treas5:
                SD_PlaySound (TREASURE2SND);
                GivePoints (2000);
                gamestate.treasurecount++;
                GetMessage("Stolen U.S. dollars: 2000 points.");
                break;
        case    bo_treas6:
                SD_PlaySound (TREASURE1SND);
                GivePoints (2400);
                gamestate.treasurecount++;
                GetMessage("Limited edition bible: 2400 points.");
                break;
        case    bo_treas7:
                SD_PlaySound (TREASURE3SND);
                GivePoints (2800);
                gamestate.treasurecount++;
                GetMessage("Gem-studded bracelet: 2800 points.");
                break;
        case    bo_treas8:
                SD_PlaySound (TREASURE4SND);
                GivePoints (3200);
                gamestate.treasurecount++;
                GetMessage("Pot of gold: 3200 points.");
                break;
        case    bo_treasnew:
                SD_PlaySound (PERCENT100SND);
                GivePoints (10000);
                gamestate.treasurecount++;
                GetMessage("Rare treasure: 10000 points!!!!!");
                break;

        case    bo_hintbub: {     // [Shaun] [tricob] hint bubbles
            GivePoints (500);
            gamestate.treasurecount++;
            ClearMemory ();
            SD_PlaySound(HINTBUBSND);
            CA_CacheGrChunk(STARTFONT + 1);
            ClearSplitVWB ();
            hintname[6] = '1'+gamestate.episode;
            memptr layout;
            if (!CA_LoadFile(hintname, &layout))
                CA_CannotOpen(hintname);
            char* hints = (char *)layout;
            char* hint = new char[255];
            lineNumber = MAPSPOT(check->tilex, check->tiley, 0) - 0x6B;
            lineNumber += (gamestate.mapon*10);
            int curLine = 0;
            int spot = 0;

            while (curLine != lineNumber) {
                if (hints[spot++] == '\n')
                    curLine++;
            }
            for (int i = 0; i < 100; i++) {
                hint[i] = hints[i + spot];
                if (hint[i] == '\n') {
                    hint[i] = '\0';
                    break;
                }
            }

            VH_UpdateScreen();
            Message(hint);
            VH_UpdateScreen();
            UNCACHEGRCHUNK(STARTFONT + 1);
            SD_WaitSoundDone();
            if (demoplayback) VW_WaitVBL(3*70);
            else
            {
                             IN_ClearKeysDown ();
                             IN_Ack ();
                             }
            free(layout);
            DrawPlayBorder ();
            break; } 

        case    bo_clip:
            if (gamestate.ammo >= AMMOMAX) return;
            SD_PlaySound (GETAMMOSND);
            GiveAmmo (AMMOCLIP);
            GetMessage("Unused clip: 8 bullets.");
            break;

        case    bo_clip2:
            if (gamestate.ammo >= AMMOMAX) return;
            SD_PlaySound (GETAMMOSND);
            GiveAmmo (AMMOCLIPUSED);
            GetMessage("Used clip: 4 bullets.");
            break;

//#ifdef SPEAR
        case    bo_25clip:
            if (gamestate.ammo >= AMMOMAX) return;
            SD_PlaySound (GETAMMOBOXSND);
            GiveAmmo (AMMOCLIP25);
            GetMessage("Ammo Box: 25 bullets.");
            break;
//#endif

        case    bo_rockets:
		if (gamestate.rockets > 7)
				       return;

            SD_PlaySound (GETAMMOBOXSND);
            GiveRockets (4);
            GetMessage("Found a box with 4 rockets.");
            break;

        case    bo_shells:
            if (gamestate.ammo2 == 200)
                return;

            SD_PlaySound (WALK1SND);
            GiveAmmo2 (16);
            GetMessage("Bullet belt: 16 Gatling gun shots.");
            break;

        case    bo_machinegun:
            SD_PlaySound (GETMACHINESND);
            GiveAmmo(6);
            GiveWeapon (wp_machinegun);
            GetMessage("Machine gun acquired.");
            break;

        case    bo_chaingun:
            SD_PlaySound (GETGATLINGSND);
            if (ffDataBottomRight!=0)GiveAmmo(6);
            if (ffDataBottomRight==0)GiveAmmo2(50);
            GiveWeapon (wp_chaingun);
            StatusDrawPic (17,4,GOTGATLINGPIC);
            facecount = 0;
            GetMessage("NOW we're cookin'!");
            break;

        case    bo_rlauncher:
            SD_PlaySound (YEAHSND);
            GiveRockets (4);
            GiveWeapon (wp_rocketgun);
            StatusDrawPic (17,4,GOTGATLINGPIC);
            facecount = 0;
            GetMessage("Rocket launcher obtained!");
            break;

        case    bo_fullheal:
            SD_PlaySound (BONUS1UPSND);
            HealSelf (MAXHEALTH);
            GiveAmmo (AMMOFULLHEAL);
            GiveExtraMan ();
            gamestate.treasurecount++;
            GetMessage("1-up obtained!");
            break;

        case    bo_food:
            if (gamestate.health >= MAXHEALTH) return;
            SD_PlaySound (HEALTH1SND);
            HealSelf (HEALTHFOOD);
            GetMessage("Dinner: 10 health points.");
            break;

        case    bo_alpo:
            if (gamestate.health >= MAXHEALTH) return;
            SD_PlaySound (HEALTH1SND);
            HealSelf (HEALTHALPO);
            GetMessage("Dog scraps: 4 health points.");
            break;

        case    bo_gibs:
            if (gamestate.health >HEALTHGIBSSTOP) return;
            SD_PlaySound (SLURPIESND);
            HealSelf (HEALTHGIBS);
            GetMessage("GIBS!!!!!!!!!!!!!");
            break;

	case    bo_poison:
		if (godmode)
		return;

		SD_PlaySound (SLURPIESND);
		TakeDamage(20,NULL);
		break;

#ifdef SPEAR
        case    bo_spear:
            spearflag = true;
            spearx = player->x;
            speary = player->y;
            spearangle = player->angle;
            playstate = ex_completed;
#endif
    }

    if (check->itemnumber != bo_hintbub) StartBonusFlash(); 
      check->shapenum = -1;      // remove from list 
}

/*
===================
=
= TryMove
=
= returns true if move ok
= debug: use pointers to optimize
===================
*/
boolean TryMove (objtype *ob)
{
    int         xl,yl,xh,yh,x,y;
    objtype    *check;
    int32_t     deltax,deltay;

    xl = (ob->x-PLAYERSIZE) >>TILESHIFT;
    yl = (ob->y-PLAYERSIZE) >>TILESHIFT;

    xh = (ob->x+PLAYERSIZE) >>TILESHIFT;
    yh = (ob->y+PLAYERSIZE) >>TILESHIFT;

#define PUSHWALLMINDIST PLAYERSIZE

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (check && !ISPOINTER(check))
            {
                if(tilemap[x][y]==255 && x==pwall.x && y==pwall.y)   // back of moving pushwall?
                {
                    switch(pwall.dir)
                    {
                        case di_north:
                            if(ob->y-PUSHWALLMINDIST<=(pwall.y<<TILESHIFT)+((63-pwall.pos)<<10))
                                return false;
                            break;
                        case di_west:
                            if(ob->x-PUSHWALLMINDIST<=(pwall.x<<TILESHIFT)+((63-pwall.pos)<<10))
                                return false;
                            break;
                        case di_east:
                            if(ob->x+PUSHWALLMINDIST>=(pwall.x<<TILESHIFT)+(pwall.pos<<10))
                                return false;
                            break;
                        case di_south:
                            if(ob->y+PUSHWALLMINDIST>=(pwall.y<<TILESHIFT)+(pwall.pos<<10))
                                return false;
                            break;
                    }
                }
                else return false;
            }
        }
    }

    //
    // check for actors
    //
    if (yl>0) yl--;
    if (yh<MAPSIZE-1) yh++;
    if (xl>0) xl--;
    if (xh<MAPSIZE-1) xh++;

    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (ISPOINTER(check) && check != player && (check->flags & FL_SHOOTABLE) )
            {
                deltax = ob->x - check->x;
#ifdef FIXACTORDISTANCE
                if (deltax < -MINPLAYERDIST || deltax > MINPLAYERDIST)
#else
                if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
#endif
                    continue;
                deltay = ob->y - check->y;
#ifdef FIXACTORDISTANCE
                if (deltay < -MINPLAYERDIST || deltay > MINPLAYERDIST)
#else
                if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
#endif
                    continue;

                return false;
            }
        }
    }
    return true;
}


/* 
===================
=
= ClipMove
=
===================
*/
void ClipMove (objtype *ob, int32_t xmove, int32_t ymove)
{
    int32_t    basex,basey;
#ifdef CHRISAUTOMAP
    DrawLevelMap();
#endif
    basex = ob->x;
    basey = ob->y;

    ob->x = basex+xmove;
    ob->y = basey+ymove;
    if (TryMove (ob)) return;

#ifndef REMDEBUG
    if (noclip && ob->x > 2*TILEGLOBAL && ob->y > 2*TILEGLOBAL
        && ob->x < (((int32_t)(mapwidth-1))<<TILESHIFT)
        && ob->y < (((int32_t)(mapheight-1))<<TILESHIFT) )
        return;         // walk through walls
#endif

#ifdef WALLSOUNDS
    if (!SD_SoundPlaying()) SD_PlaySound (HITWALLSND);
#endif

    ob->x = basex+xmove;
    ob->y = basey;
    if (TryMove (ob)) return;

    ob->x = basex;
    ob->y = basey+ymove;
    if (TryMove (ob)) return;

    ob->x = basex;
    ob->y = basey;
}

//==========================================================================

//#ifdef WARPING
void ResetPlayer (void)
{
	buttonstate[bt_attack] = false;
	buttonheld[bt_attack] = false;
	gamestate.attackframe = gamestate.weaponframe = 0;
//	DrawPlayerWeapon();
//	ThreeDRefresh();
}

#ifdef WARPING
void WarpPlayer (int facedir)
{
	objtype  *check;
	int  x,y;

	ResetPlayer ();
	switch (facedir)
	{
		case EAST:
 			check = actorat[warpex][warpey];
 			x = warpex; y = warpey;
  			if (gamestate.warped == 4) gamestate.warped = 0; else gamestate.warped = 3;
  			break;
		case WEST:
  			check = actorat[warpwx][warpwy];
  			x = warpwx; y = warpwy;
  			if (gamestate.warped == 3) gamestate.warped = 0; else gamestate.warped = 4;
  			break;
		case NORTH:
			check = actorat[warpnx][warpny];
  			x = warpnx; y = warpny;			
  			if (gamestate.warped == 2) gamestate.warped = 0; else gamestate.warped = 1;    			
  			break;
		case SOUTH:
  			check = actorat[warpsx][warpsy];
  			x = warpsx; y = warpsy;  			
  			if (gamestate.warped == 1) gamestate.warped = 0; else gamestate.warped = 2;
  			break;
  	}

	if (check > objlist && (check->flags & FL_SHOOTABLE)) KillActor(check);
	SpawnPlayer(x,y,facedir);

  		
	SD_PlaySound (BOSSACTIVESND);
	if (DigiMode != sds_Off)
  	{
  		long lasttimecount = TimeCount;
  		while(TimeCount < lasttimecount+35) //adjust warp duration here
   		SD_Poll();
  	}
	else
  		VW_WaitVBL(50);

	ResetPlayer ();
}
#endif

/*
===================
=
= VictoryTile
=
===================
*/

void VictoryTile (void)
{

#ifdef TEKZOOMSHOCK   // BEGIN ZOOM BLOCK 
   ResetZoomEffects(); 
#endif                // END ZOOM BLOCK 

#ifndef SPEAR
    SpawnBJVictory ();
#endif

    gamestate.victoryflag = true;
}

/*
===================
=
= Thrust
=
===================
*/

// For player movement in demos exactly as in the original Wolf3D v1.4 source code
static fixed FixedByFracOrig(fixed a, fixed b)
{
    int sign = 0;
    if(b == 65536) b = 65535;
    else if(b == -65536) b = 65535, sign = 1;
    else if(b < 0) b = (-b), sign = 1;

    if(a < 0)
    {
        a = -a;
        sign = !sign;
    }
    fixed res = (fixed)(((int64_t) a * b) >> 16);
    if(sign)
        res = -res;
    return res;
}

void Thrust (int angle, int32_t speed)
{
    int32_t xmove,ymove;
    unsigned offset;


    //
    // ZERO FUNNY COUNTER IF MOVED!
    //
#ifdef SPEAR
    if (speed)
        funnyticount = 0;
#endif

    thrustspeed += speed;
    //
    // moving bounds speed
    //
    if (speed >= MINDIST*2)
        speed = MINDIST*2-1;

    xmove = DEMOCHOOSE_ORIG_SDL(
                FixedByFracOrig(speed, costable[angle]),
                FixedMul(speed,costable[angle]));
    ymove = DEMOCHOOSE_ORIG_SDL(
                -FixedByFracOrig(speed, sintable[angle]),
                -FixedMul(speed,sintable[angle]));

    ClipMove(player,xmove,ymove);

    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    offset = (player->tiley<<mapshift)+player->tilex;
    player->areanumber = *(mapsegs[0] + offset) -AREATILE;

    if (*(mapsegs[1] + offset) != gamestate.teleportnum) gamestate.teleportnum = 0;

    switch (*(mapsegs[1] + offset))
    {
// BJ "picks up" lightning markers and guard block 1
      case BLOCKGUARD1:
      { *(mapsegs[1]+((player->tiley<<mapshift))+(player->tilex)) = 0; break; }
      case STARTRAIN: {
      *(mapsegs[1]+((player->tiley<<mapshift))+(player->tilex)) = 0;
      StartBonusFlash();
      NoShadingStuff = true;
      SD_PlaySound (THUNDEROUTDOORSND);
      break; }
      case EXITTILE: {
      VictoryTile ();
      break; }
      case ENDLEVELTILE: {
           gamestate.attackframe = gamestate.weaponframe = 0;
           SD_PlaySound (YEAHSND);
           playstate = ex_completed;
           break; }

      case TELELADDER1A:
      case TELELADDER1B:
      case TELELADDER2A:
      case TELELADDER2B:
      case TELEPORT1A:
      case TELEPORT1B:
      case TELEPORT2A:
      case TELEPORT2B:
      case TELEPORT3A:
      case TELEPORT3B:
      case TELEPORT4A:
      case TELEPORT4B: {
           int BJsSquare = *(mapsegs[1] + offset);
           if (BJsSquare != gamestate.teleportnum) {
           switch (BJsSquare) {
                  case TELELADDER1A:
                       T_Teleport (TeleLaddX1B,TeleLaddY1B);
                       break;
                  case TELELADDER1B:
                       T_Teleport (TeleLaddX1A,TeleLaddY1A);
                       break;
                  case TELELADDER2A:
                       T_Teleport (TeleLaddX2B,TeleLaddY2B);
                       break;
                  case TELELADDER2B:
                       T_Teleport (TeleLaddX2A,TeleLaddY2A);
                       break;
                  case TELEPORT1A:
                       T_Teleport (TeleportX1B,TeleportY1B);
                       break;
                  case TELEPORT1B:
                       T_Teleport (TeleportX1A,TeleportY1A);
                       break;
                  case TELEPORT2A:
                       T_Teleport (TeleportX2B,TeleportY2B);
                       break;
                  case TELEPORT2B:
                       T_Teleport (TeleportX2A,TeleportY2A);
                       break;
                  case TELEPORT3A:
                       T_Teleport (TeleportX3B,TeleportY3B);
                       break;
                  case TELEPORT3B:
                       T_Teleport (TeleportX3A,TeleportY3A);
                       break;
                  case TELEPORT4A:
                       T_Teleport (TeleportX4B,TeleportY4B);
                       break;
                  case TELEPORT4B:
                       T_Teleport (TeleportX4A,TeleportY4A);
                       break;
                       }                       
                       break; } break; }

#ifdef WARPING
      case WARPEASTTILE: { WarpPlayer(EAST); break; }
      case WARPWESTTILE: { WarpPlayer(WEST); break; }
      case WARPNORTHTILE:{ WarpPlayer(NORTH); break; }
      case WARPSOUTHTILE:{ WarpPlayer(SOUTH); break; }
#endif
      default: break;
    }
}


/*
=============================================================================

                                ACTIONS

=============================================================================
*/


/*
===============
=
= Cmd_Fire
=
===============
*/

void Cmd_Fire (void)
{
    buttonheld[bt_attack] = true;

    gamestate.weaponframe = 0;

    player->state = &s_attack;

    gamestate.attackframe = 0;
    gamestate.attackcount =
        attackinfo[gamestate.weapon][gamestate.attackframe].tics;
    gamestate.weaponframe =
        attackinfo[gamestate.weapon][gamestate.attackframe].frame;
}

//===========================================================================

/*
===============
=
= Cmd_Use
=
===============
*/

void Cmd_Use (void)
{
    int     checkx,checky,doornum,dir,tileobj;
    boolean elevatorok;
    statobj_t      *statptr;

    dir = BJ_Facing ();
    elevatorok = false;
    switch (dir)
    {
      case di_east:
      case di_west: elevatorok = true;
      default: break;
    }
    checkx = player->tilex + dx4dir[dir];
    checky = player->tiley + dy4dir[dir];

    tileobj = MAPSPOT(checkx,checky,1);
    doornum = tilemap[checkx][checky];
    if (tileobj == PUSHABLETILE 
#ifdef WALLPATCHES      
      || ((tileobj >= PFIRSTPATCH) && (tileobj < PLASTPATCH))
#endif
       ) // pushable wall
    {
        PushWall (checkx,checky,dir);
        return;
    }

// Interactive Objects code, from BT's usable objects source
    for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
    {
            if(statptr->tilex == checkx && statptr->tiley == checky && !buttonheld[bt_use])
            {
                buttonheld[bt_use] = true;
                if(statptr->shapenum == AMMO_CRATE && gamestate.ammo2 != 200)
                {
                                        statptr->shapenum = CRAT_EMPTY;
                                        SD_PlaySound (WALK2SND);
                                        GiveAmmo2(75);
                                        GetMessage("75 Gatling Gun shots obtained!");
                                        StartBonusFlash ();
                                        break;
                }
                if(statptr->shapenum == RECT_BBELT && gamestate.ammo2 != 200)
                {
                                        statptr->shapenum = RECT_EMPTY;
                                        SD_PlaySound (WALK1SND);
                                        GiveAmmo2(15);
                                        GetMessage("15 Gatling Gun shots obtained.");
                                        StartBonusFlash ();
                                        break;
                }
                if(statptr->shapenum == RECT_GAGUN)
                {
                         statptr->shapenum = RECT_EMPTY;
                         SD_PlaySound (GETGATLINGSND);
                         if (ffDataBottomRight)GiveAmmo(6);
                         if (!ffDataBottomRight)GiveAmmo2(50);
                         GiveWeapon (wp_chaingun);
                         StatusDrawPic (17,4,GOTGATLINGPIC);
                         facecount = 0;
                         GetMessage("Gatling gun!!!!!!!");
                         StartBonusFlash ();
                         break;
                }
                if(statptr->shapenum == TABL_RLAUNCH)
                {
                         statptr->shapenum = RECT_EMPTY;
                         SD_PlaySound (YEAHSND);
                         GiveRockets (4);
                         GiveWeapon (wp_rocketgun);
                         StatusDrawPic (17,4,GOTGATLINGPIC);
                         facecount = 0;
                         GetMessage("I'm a rocket man!");
                         StartBonusFlash ();
                         break;
                }
                if(statptr->shapenum == BOMB_CRATE)
                {
                        statptr->shapenum = CRAT_EMPTY;
                        SD_PlaySound (TAKEDAMAGESND);
                        TakeDamage (30,NULL);
                        GetMessage("Booby-trapped!");
                        break;
                }

                switch (statptr->shapenum)
                {
                    case TABL_KEY3:
                    case TABL_KEY4:
                {
                        GiveKey (statptr->shapenum-SPR_CHEAT);
                        statptr->shapenum = RECT_EMPTY;
                        SD_PlaySound (GETKEYSND);
                        GetMessage("Got a key from a table.");
                        StartBonusFlash ();
                        break;
                }
                    case TABL_TREAS1:
                        SD_PlaySound (BONUS1SND);
                        GivePoints (BONUSPOINT1);
                        gamestate.treasurecount++;
                        statptr->shapenum = TABL_EMPTY;
                        StartBonusFlash ();
                        GetMessage("Cross: 100 points.");
                        break;
                    case TABL_TREAS2:
                        SD_PlaySound (BONUS2SND);
                        GivePoints (BONUSPOINT2);
                        gamestate.treasurecount++;
                        statptr->shapenum = TABL_EMPTY;
                        GetMessage("Chalice: 500 points.");
                        StartBonusFlash ();
                        break;
                    case TABL_TREAS3:
                        SD_PlaySound (BONUS3SND);
                        GivePoints (BONUSPOINT3);
                        gamestate.treasurecount++;
                        statptr->shapenum = TABL_EMPTY;
                        GetMessage("Chest of jewels: 1000 points.");
                        StartBonusFlash ();
                        break;
                    case TABL_TREAS4:
                        SD_PlaySound (BONUS4SND);
                        GivePoints (BONUSPOINT4);
                        gamestate.treasurecount++;
                        statptr->shapenum = TABL_EMPTY;
                        GetMessage("Crown: 5000 points!");
                        StartBonusFlash ();
                        break;

                    case TABL_ITEM1:
                    case TABL_ITEM2:
                        GiveKey (statptr->shapenum - TABL_ITEM1);
                        SD_PlaySound (GETKEYSND);
                        GetMessage("Got a key from a table.");
                        statptr->shapenum = TABL_EMPTY;
                        StartBonusFlash ();
                        break;

                    case TABL_ITEM3:
                        if(gamestate.health == 100)
                        return;
                        HealSelf (10);
                        SD_PlaySound (HEALTH1SND);
                        GetMessage("Got food from a table.");
                        statptr->shapenum = TABL_EMPTY;
                        StartBonusFlash ();
                        break;

                    case TABL_ITEM4:
                        if(gamestate.health == 100)
                        return;
                        HealSelf (25);
                        SD_PlaySound (HEALTH2SND);
                        GetMessage("Got a medkit from a table.");
                        statptr->shapenum = TABL_EMPTY;
                        StartBonusFlash ();
                        break;

                    case TABL_ANTID:
                        if(gamestate.antidote == true)
                        return;
                        gamestate.antidote = true;
                        SD_PlaySound (ANTIDOTESND);
                        GetMessage("Got the antidote from a table.");
                        statptr->shapenum = RECT_EMPTY;
                        StartBonusFlash ();
                        break;

                    case TABL_ITEM5:
                        if(gamestate.ammo == 99)
                        return;
                        GiveAmmo(25);
                        SD_PlaySound (GETAMMOBOXSND);
                        GetMessage("Got 25 ammo from a table.");
                        statptr->shapenum = TABL_EMPTY;
                        StartBonusFlash ();
                        break;
                    case TABL_ANIM0:
                        SD_PlaySound (ATKKNIFESND);
                        GetMessage("Unlit the candle.");
                        statptr->flags = 0;
                        statptr->shapenum = SPR_STAT_2;
                        break;
                    case SPR_STAT_2:
                        SD_PlaySound (ESCPRESSEDSND);
                        GetMessage("Lit the candle.");
                        statptr->flags = FL_FULLBRIGHT;
                        statptr->shapenum = TABL_ANIM0;
                        break;
                }
            }
        }
// End Interactive Objects code

    if (!buttonheld[bt_use] && doornum == ELEVATORTILE && elevatorok) // use elevator
    {
        buttonheld[bt_use] = true;

        tilemap[checkx][checky]++;              // flip switch
        if (*(mapsegs[0]+(player->tiley<<mapshift)+player->tilex) == ALTELEVATORTILE)
            playstate = ex_secretlevel;
        else
            playstate = ex_completed;
        SD_PlaySound (LEVELDONESND);
        SD_WaitSoundDone();
    }
    else if (!buttonheld[bt_use] && doornum & 0x80)
    {
        buttonheld[bt_use] = true;
        OperateDoor (doornum & ~0x80);
    }
#ifdef WALLSOUNDS
    else
        SD_PlaySound (DONOTHINGSND);
#endif
}

/*
=============================================================================

                                PLAYER CONTROL

=============================================================================
*/



/*
===============
=
= SpawnPlayer
=
===============
*/

void SpawnPlayer (int tilex, int tiley, int dir)
{
    int temp,i;

    gamestate.bjthrust = 0;
    player->obclass = playerobj;
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    player->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    player->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    player->state = &s_player;
    i = (1-dir)*90;
    if (dir == NULLDIR) i = temp;
    player->angle = i;
    if (player->angle<0) player->angle += ANGLES;
    player->flags = FL_NEVERMARK;
    Thrust (0,0);                           // set some variables

    InitAreas ();
}


//===========================================================================

//===========================================================================

void RocketLauncherAttack (objtype *ob)
{
   gamestate.rockets--;
   DrawAmmo ();
   SD_PlaySound (MISSILEFIRESND);
   
   GetNewActor ();

   newobj->state = &s_rocket;
   newobj->ticcount = 1;

    newobj->x = ob->x ;
    newobj->y = ob->y ;
    newobj->tilex = newobj->x >> TILESHIFT;
    newobj->tiley = newobj->y >> TILESHIFT;

   newobj->obclass = rocketobj;
   newobj->dir = nodir;
   newobj->angle = ob->angle;
   newobj->speed = 0x4200l;
   newobj->flags = FL_NONMARK|FL_BONUS;
   newobj->active = ac_yes;
}

/*
===============
=
= T_KnifeAttack
=
= Update player hands, and try to do damage when the proper frame is reached
=
===============
*/

void    KnifeAttack (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;

    SD_PlaySound (ATKKNIFESND);
    // actually fire
    dist = 0x7fffffff;
    closest = NULL;
    for (check=ob->next; check; check=check->next)
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && abs(check->viewx-centerx) < shootdelta)
        {
            if (check->transx < dist)
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (!closest || dist > 0x18000l)
    {
        // missed
        return;
    }

    // hit something
    DamageActor (closest,(US_RndT() >> 4) + 1);
}


void    GunAttack (objtype *ob)
{
    objtype *check,*closest,*oldclosest;
    int      damage;
    int      dx,dy,dist;
    int32_t  viewdist;

if (gamestate.weapon == wp_rocketgun)
   {
      madenoise = true;
      RocketLauncherAttack (ob);
      return;
        }

    switch (gamestate.weapon)
    {
        case wp_pistol:
	if (!ffDataBottomRight) SD_PlaySound (NEWWPN2SND);
       else SD_PlaySound (ATKPISTOLSND);
            break;
        case wp_machinegun:
            if (!ffDataBottomRight) SD_PlaySound (NEWWPN3SND);
            else SD_PlaySound (ATKMACHINEGUNSND);
            break;
        case wp_chaingun:
            if (!ffDataBottomRight) SD_PlaySound (MAGNUMFIRESND);
            else SD_PlaySound (ATKGATLINGSND);
            break;
    }

    madenoise = true;

#ifdef TEKZOOMSHOCK     // BEGIN ZOOM BLOCK 
    StartRecoilEffect( 50 ); 
#endif                  // END ZOOM BLOCK 
    //
    // find potential targets
    //
    viewdist = 0x7fffffffl;
    closest = NULL;

    while (1)
    {
        oldclosest = closest;

        for (check=ob->next ; check ; check=check->next)
        {
            if ((check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
                && abs(check->viewx-centerx) < shootdelta)
            {
                if ( (check->transx < viewdist) && (check->obclass != madmeddormantobj) ) // Don't hit idle Mad Medkits
                {
                    viewdist = check->transx;
                    closest = check;
                }
            }
        }

        if (closest == oldclosest)
            return;                                         // no more targets, all missed

        //
        // trace a line from player to enemey
        //
        if (CheckLine(closest))
            break;
    }

    //
    // hit something
    //
    dx = ABS(closest->tilex - player->tilex);
    dy = ABS(closest->tiley - player->tiley);
    dist = dx>dy ? dx:dy;
#ifdef HIRES_GUNFIX
    if (dist<4)
#else
    if (dist<2)
#endif
        damage = US_RndT() / 4;
#ifdef HIRES_GUNFIX
    else if (dist<8)
#else
    else if (dist<4)
#endif
        damage = US_RndT() / 6;
    else
    {
#ifdef HIRES_GUNFIX
        if ( (US_RndT() / 6) < dist)           // missed
#else
        if ( (US_RndT() / 12) < dist)           // missed
#endif
            return;
        damage = US_RndT() / 6;
    }
    DamageActor (closest,damage);
}

//===========================================================================

/*
===============
=
= VictorySpin
=
===============
*/

void VictorySpin (void)
{
    int32_t    desty;

    if (player->angle > 270)
    {
        player->angle -= (short)(tics * 3);
        if (player->angle < 270)
            player->angle = 270;
    }
    else if (player->angle < 270)
    {
        player->angle += (short)(tics * 3);
        if (player->angle > 270)
            player->angle = 270;
    }

    desty = (((int32_t)player->tiley-5)<<TILESHIFT)-0x3000;

    if (player->y > desty)
    {
        player->y -= tics*4096;
        if (player->y < desty)
            player->y = desty;
    }
}


//===========================================================================

/*
===============
=
= T_Attack
=
===============
*/

void    T_Attack (objtype *ob)
{
    int             temp,temp2;
    atkinf*         temp3;
    struct  atkinf  *cur;

    UpdateFace ();

    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        buttonstate[bt_use] = false;

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        buttonstate[bt_attack] = false;

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    //
    // change frame and fire
    //
    gamestate.attackcount -= (short) tics;
    while (gamestate.attackcount <= 0)
    {
        cur = &attackinfo[gamestate.weapon][gamestate.attackframe];
        switch (cur->attack)
        {
            case -1:
                ob->state = &s_player;
                if (ffDataBottomRight!=0 && !gamestate.ammo)
                {
                    gamestate.weapon = wp_knife;
                    DrawWeapon ();
                }
                else
                {
                    if (gamestate.weapon != gamestate.chosenweapon)
                    {
                        gamestate.weapon = gamestate.chosenweapon;
                        DrawWeapon ();
                    }
                }
                gamestate.attackframe = gamestate.weaponframe = 0;
                return;

            case 4:
            if (ffDataBottomRight!=0 && !gamestate.ammo) break;
            if (gamestate.weapon == wp_rocketgun && !gamestate.rockets) break;
            if (ffDataBottomRight==0 && !gamestate.ammo2) break;
                if (buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
            case 1:
                if ((ffDataBottomRight!=0 && !gamestate.ammo)||(ffDataBottomRight==0 && !gamestate.ammo2 && gamestate.weapon == wp_chaingun)||(!gamestate.ammo && gamestate.weapon != wp_chaingun))
                {       // can only happen with chain gun
                    gamestate.attackframe++;
                    break;
                }
                GunAttack (ob);
                if (!ammocheat)
			{
            if (ffDataBottomRight==0 && gamestate.weapon == wp_chaingun) gamestate.ammo2--;
            if (gamestate.weapon != wp_rocketgun && (ffDataBottomRight!=0 || gamestate.weapon != wp_chaingun) ) gamestate.ammo--;
			}
                DrawAmmo ();
                break;

            case 2:
                KnifeAttack (ob);
                break;

            case 3:
                if (gamestate.ammo && buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
                break;
#ifdef NEWGUNCODE
            case 99:
                if((!ffDataBottomRight && gamestate.weapon == wp_chaingun && !gamestate.ammo2) || (gamestate.weapon != wp_rocketgun && gamestate.weapon != wp_chaingun && !gamestate.ammo) || (gamestate.weapon == wp_rocketgun && !gamestate.rockets) || (ffDataBottomRight && gamestate.weapon == wp_chaingun && !gamestate.ammo) ) {
                    gamestate.attackcount = 6;
                    gamestate.attackframe = 4;
                    gamestate.weaponframe = -4;
                    ob->state = &s_player;
                    GetMessage("You don't have ammo for this weapon.");
                    SD_PlaySound (ESCPRESSEDSND); }
                    break;
#endif
        }

        gamestate.attackcount += cur->tics;
        gamestate.attackframe++;
        gamestate.weaponframe =
            attackinfo[gamestate.weapon][gamestate.attackframe].frame;
    }
}



//===========================================================================

/*
===============
=
= T_Player
=
===============
*/

void    T_Player (objtype *ob)
{
    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }
#ifdef CHRISAUTOMAP
    DrawLevelMap();
#endif
#ifdef STATBARCLOCK
	  DrawTime ();
#endif
    UpdateFace ();
    CheckWeaponChange ();

    if ( buttonstate[bt_use] )
        Cmd_Use ();

    if (buttonstate[bt_attack] && !buttonheld[bt_attack])
#ifndef NEWGUNCODE
    {
    if((ffDataBottomRight==0 && gamestate.weapon == wp_chaingun && gamestate.ammo2) || (gamestate.weapon == wp_knife) || (gamestate.weapon != wp_rocketgun && gamestate.weapon != wp_chaingun && gamestate.ammo) || (gamestate.weapon == wp_rocketgun && gamestate.rockets) || (ffDataBottomRight!=0 && gamestate.weapon == wp_chaingun && gamestate.ammo) )
    {
#endif
        Cmd_Fire ();
#ifndef NEWGUNCODE
        }
        else
        {
        GetMessage("You don't have ammo for this weapon.");
        SD_PlaySound (ESCPRESSEDSND);    
        }
    }
#endif

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);
}

void T_Teleport (int tilex, int tiley)
{
    int temp;

    VW_FadeOut ();
    player->obclass = playerobj;
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    player->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    player->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    player->state = &s_player;
    player->flags = FL_NEVERMARK;
#ifdef FFTWOPARTS
    if (!(MAPSPOT(player->tilex,player->tiley,2)>>8)) gamestate.outside = true;
#endif
    gamestate.attackframe = gamestate.attackcount = gamestate.weaponframe = 0;
    VW_WaitVBL (140); // 70 * 2 (seconds) = 140
    IN_ClearKeysDown ();
    gamestate.teleportnum = MAPSPOT (tilex, tiley, 1);
    DrawPlayBorder ();
    VW_UpdateScreen ();
    Thrust (0,0);                           // set some variables
    InitAreas ();
}
