#ifndef WL_DEF_H
#define WL_DEF_H
/* ==================================================================
                      WL_DEF.H - Definitions File
   ==================================================================
       Please Do Not Modify Values in Any of the files unless you
       know what you are doing.  Any game specific settings have
       been moved to a file called 'GAMESETTINGS.H'.
   ================================================================== */
// Defines which version shall be built and configures supported extra features
#include "version.h"

#include <assert.h>
#include <fcntl.h>
#include <math.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#if defined(_arch_dreamcast)
#	include <kos.h>
#elif !defined(_WIN32)
#	include <stdint.h>
#	include <string.h>
#	include <stdarg.h>
#endif
#include <SDL.h>

#if !defined O_BINARY
#	define O_BINARY 0
#endif

#pragma pack(1)

#include "foreign.h"

#ifndef SPEAR
    #include "audiowl6.h"
    #ifdef UPLOAD
        #include "gfxv_apo.h"
    #else
        #ifdef GOODTIMES
            #include "gfxv_wl6.h"
        #else
            #include "gfxv_apo.h"
        #endif
    #endif
#else
    #include "audiosod.h"
    #include "gfxv_sod.h"
    #include "f_spear.h"
#endif

#ifdef _arch_dreamcast
typedef uint8 uint8_t;
typedef uint16 uint16_t;
typedef uint32 uint32_t;
typedef int8 int8_t;
typedef int16 int16_t;
typedef int32 int32_t;
typedef int64 int64_t;
typedef ptr_t uintptr_t;
#endif

typedef uint8_t byte;
typedef uint16_t word;
typedef int32_t fixed;
typedef uint32_t longword;
typedef int8_t boolean;
typedef void * memptr;

typedef struct
{
    int x,y;
} Point;
typedef struct
{
    Point ul,lr;
} Rect;

void Quit(const char *errorStr, ...);

#include "id_pm.h"
#include "id_sd.h"
#include "id_in.h"
#include "id_vl.h"
#include "id_vh.h"
#include "id_us.h"
#include "id_ca.h"

#include "wl_menu.h"

#include "wl_sprites.h"				// Sprites Listing

#include "wl_gamesettings.h"					// Game Settings - Easier to Configure game to liking

/* =============================================================================
                            GLOBAL CONSTANTS
   ============================================================================= */

#define MAXTICS										 10
#define DEMOTICS										4
#undef	M_PI
#define PI													3.14159265358979323846264338327950288419716939937510
#define M_PI PI
#define NUMBERCHARS									9
#define SCREENSEG										0xa000
#define SCREENBWIDE								 80
#define HEIGHTRATIO									0.50            // also defined in id_mm.c

/* ===================================================================
                            MACROS
   =================================================================== */
//#define MAPSPOT(x,y,plane)         (mapsegs[plane]+y*MAPSIZE+x)
#define MAPSPOT(x,y,plane)				 (mapsegs[plane][((y)<<mapshift)+(x)])
#define SIGN(x)										 ((x)>0?1:-1)
#define ABS(x)										 ((int)(x)>0?(x):-(x))
#define LABS(x)										 ((int32_t)(x)>0?(x):-(x))
#define abs(x)										 ABS(x)
#define min_of_2(x, y)						 ((x)<(y)?(x):(y))
#define max_of_2(x, y)						 ((x)>(y)?(x):(y))
#define DEG2RAD(x)								 (((x)*M_PI)/180.0)
#define RAD2DEG(x)								 (((x)*180.0)/M_PI)
#define SHIFTR(x,y)                ((x)>>y)
#define SHIFTL(x,y)                ((x)<<y)
#define SHIFTR4(x)                 ((x)>>4)
#define SHIFTL4(x)                 ((x)<<4)
#define SHIFTR8(x)                 ((x)>>8)
#define SHIFTL8(x)                 ((x)<<8)
#define TILESHIFTR(x)              ((x)>>TILESHIFT)
#define TILESHIFTL(x)              ((x)<<TILESHIFT)
#define TILE2POS(x)                (TILESHIFTL((int32_t)x) + TILEGLOBAL/2)
#define POS2UPOS(x)                ((x)>>UNSIGNEDSHIFT)
#define FIX_HEIGHT(pix_height)     ((pix_height)<<2)
#define PIX_HEIGHT(fix_height)     (((fixheight)>>2)&~1)
#define PIX_HALFHEIGHT(fix_height) ((fixheight)>>3)
#define MAKEANIM(highbyte,lowbyte) (byte)((SHIFTL4(highbyte&0xf)) + (lowbyte&0xf))


#define SEAMLESSSECRETSTART        SEAMLESSSTART
#define SEAMLESSSECRETEND          (SEAMLESSSTART + SEAMLESSSECRET)

#ifndef SPEAR
  #ifndef SEAMLESS
    #define LRpack									8         // # of levels to store in endgame
  #else  
    #define LRpack                  (NUMMAPS - SEAMLESSSECRET)
  #endif
#else
    #define LRpack								 20
#endif



#define DOORWALL          (PMSpriteStart - (NUMDOORPICS * 2))     
#define NUMPATCHES        (NUMWALLPATCHPICS * 2)
#define PATCHWALL         (DOORWALL - NUMPATCHES)
#define FIRSTPATCH        FIRSTPATCHIDENTIFIER
#define LASTPATCH         (FIRSTPATCH + NUMPATCHES) 
#define PFIRSTPATCH       PUSHPATCHIDENTIFIER
#define PLASTPATCH        (PUSHPATCHIDENTIFIER + NUMPATCHES)
#define PUSHPATCHDIFF     (PFIRSTPATCH - FIRSTPATCH)

#define JAWSDIST						  (0x7200l)     // Radius of Jaws's wrath!

#define MINDIST						  (0x5800l)
#define PLAYERSIZE						MINDIST         // player radius
#define MINPLAYERDIST 0xd800l // MCS: used in WL_AGENT.C instead of MINACTORDIST
                              // so player can pass between fake boss objects

#define MINACTORDIST					0x10000l        // minimum dist from player center
                                                        // to any actor center
#define GLOBAL1								(1l<<16)
#define TILEGLOBAL						GLOBAL1
#define PIXGLOBAL					(GLOBAL1/64)
#define TILESHIFT								  16l
#define UNSIGNEDSHIFT								8

#define ANGLES										360         // must be divisable by 4
#define ANGLEQUAD					 (ANGLES/4)
#define FINEANGLES							 3600
#define ANG90          (FINEANGLES/4)
#define ANG180						  (ANG90*2)
#define ANG270						  (ANG90*3)
#define ANG360							(ANG90*4)
#define VANG90						 (ANGLES/4)
#define VANG180						 (VANG90*2)
#define VANG270						 (VANG90*3)
#define VANG360						 (VANG90*4)

#define mapshift										6
#define MAPSIZE         (1<<mapshift)
#define maparea       MAPSIZE*MAPSIZE

#define mapheight				      MAPSIZE
#define mapwidth							MAPSIZE

#ifdef USE_HIRES
	#define		TEXTURESHIFT						7
	#define		TEXTURESIZE							(1<<TEXTURESHIFT)
	#define		TEXTUREFROMFIXEDSHIFT		2
	#define		TEXTUREMASK							(TEXTURESIZE*(TEXTURESIZE-1))
	#define		SPRITESCALEFACTOR				1

	#ifdef USE_RAYCASTOR
		#define RESVAL1									2
		#define RESVAL2									0x3F80
		#define RESVAL3								127
		#define RESVAL4								 64
		#define RESVAL5									9
		#define RESVAL6									7
		#define RESVAL7									1
		#define RESVAL8								129
		#define RESVAL9								128
		#define RESVAL10							 17
		#define RESVAL11						16384
		#define RESVAL12						   11
		#define RESVAL13							  6
		#define RESVAL14						  256
		#define RESVAL15							  8
		#define RESVAL16								3
	#endif
#else
	#define		TEXTURESHIFT						6
	#define		TEXTURESIZE							(1<<TEXTURESHIFT)
	#define		TEXTUREFROMFIXEDSHIFT		4
	#define		TEXTUREMASK							(TEXTURESIZE*(TEXTURESIZE-1))
	#define		SPRITESCALEFACTOR				2

	#ifdef USE_RAYCASTOR
		#define RESVAL1									4
		#define RESVAL2									0xFC0
		#define RESVAL3								 63
		#define RESVAL4								 32
		#define RESVAL5								 10
		#define RESVAL6									6
		#define RESVAL7									2
		#define RESVAL8								 65
		#define RESVAL9								 64
		#define RESVAL10							 16
		#define RESVAL11						 4096
		#define RESVAL12							 10
		#define RESVAL13								5
		#define RESVAL14							127
		#define RESVAL15								7
		#define RESVAL16								5
	#endif
#endif

#define NORTH   0
#define EAST    1
#define SOUTH   2
#define WEST    3
#define NULLDIR 4

#define SCREENSIZE      (SCREENBWIDE*208)
#define PAGE1START      0
#define PAGE2START      (SCREENSIZE)
#define PAGE3START      (SCREENSIZE*2u)
#define FREESTART       (SCREENSIZE*3u)

#define PIXRADIUS       512


// object flag values

typedef enum
{
    FL_SHOOTABLE        = 0x00000001,
    FL_BONUS            = 0x00000002,
    FL_NEVERMARK        = 0x00000004,
    FL_VISABLE          = 0x00000008,
    FL_ATTACKMODE       = 0x00000010,
    FL_FIRSTATTACK      = 0x00000020,
    FL_AMBUSH           = 0x00000040,
    FL_NONMARK          = 0x00000080,
    FL_FULLBRIGHT       = 0x00000100,
#ifdef USE_DIR3DSPR
    // you can choose one of the following values in wl_act1.cpp
    // to make a static sprite a directional 3d sprite
    // (see example at the end of the statinfo array)
    FL_DIR_HORIZ_MID    = 0x00000200,
    FL_DIR_HORIZ_FW     = 0x00000400,
    FL_DIR_HORIZ_BW     = 0x00000600,
    FL_DIR_VERT_MID     = 0x00000a00,
    FL_DIR_VERT_FW      = 0x00000c00,
    FL_DIR_VERT_BW      = 0x00000e00,

    // these values are just used to improve readability of code
    FL_DIR_NONE         = 0x00000000,
    FL_DIR_POS_MID      = 0x00000200,
    FL_DIR_POS_FW       = 0x00000400,
    FL_DIR_POS_BW       = 0x00000600,
    FL_DIR_POS_MASK     = 0x00000600,
    FL_DIR_VERT_FLAG    = 0x00000800,
    FL_DIR_MASK         = 0x00000e00,
#endif
    // next free bit is   0x00001000
} objflag_t;

/*
=============================================================================

                               GLOBAL TYPES

=============================================================================
*/

typedef enum {
    di_north,
    di_east,
    di_south,
    di_west
} controldir_t;

typedef enum {
    dr_normal,
    dr_lock1,
    dr_lock2,
    dr_lock3,
    dr_lock4,
    dr_elevator,
    dr_normal2,
    dr_normal3,
    dr_types
} door_t;

typedef enum {
    ac_badobject = -1,
    ac_no,
    ac_yes,
    ac_allways
} activetype;

typedef enum {
    nothing,
    playerobj,
    inertobj,
    guardobj,
    darrenobj,
    officerobj,
    ssobj,
    dogobj,
#ifdef USERATS
    ratobj,
#endif
    jawsobj,
    bossobj,
    schabbobj,
    fakeobj,
    mechahitlerobj,
    mutantobj,
    needleobj,
    fireobj,
    bjobj,
    ghostobj,
    realhitlerobj,
    gretelobj,
    giftobj,
    fatobj,
    rocketobj,

    spectreobj,
    angelobj,
    transobj,
    uberobj,
    willobj,
    deathobj,
    hrocketobj,
    sparkobj,

    gatlerobj,
    madmeddormantobj,
    madmedobj,
    barrel1obj,
    barrel2obj,
    plant1obj,
    plant2obj,
    plant3obj,
    plant4obj,
    purplebloodobj,
} classtype;

typedef enum {
    none,
    block,
    bo_tabletreas,
    bo_gibs,
    bo_alpo,
    bo_firstaid,
    bo_key1,
    bo_key2,
    bo_key3,
    bo_key4,
    bo_cross,
    bo_chalice,
    bo_bible,
    bo_crown,
    bo_treas1,
    bo_treas2,
    bo_treas3,
    bo_treas4,
    bo_treas5,
    bo_treas6,
    bo_treas7,
    bo_treas8,
    bo_treasnew,
    bo_hintbub,
    bo_clip,
    bo_clip2,
    bo_shells,
    bo_machinegun,
    bo_chaingun,
    bo_food,
    bo_fullheal,
    bo_25clip,
    bo_spear,
    bo_poison,
    bo_portmed,
    bo_rlauncher,
    bo_rockets,
    bo_detect,

} wl_stat_t;

typedef enum {
    east,
    northeast,
    north,
    northwest,
    west,
    southwest,
    south,
    southeast,
    nodir
} dirtype;

typedef enum {
    en_guard,
    en_darren,
    en_officer,
    en_ss,
    en_dog,
#ifdef USERATS
    en_rat,
#endif
    en_jaws,
    en_boss,
    en_schabbs,
    en_fake,
    en_hitler,
    en_mutant,
    en_blinky,
    en_clyde,
    en_pinky,
    en_inky,
    en_gretel,
    en_gift,
    en_fat,
    en_spectre,
    en_angel,
    en_trans,
    en_uber,
    en_will,
    en_death,
    en_gatler,
    en_barrel1,
    en_barrel2,
    en_plant1,
    en_plant2,
    en_plant3,
    en_plant4,
    NUMENEMIES
} enemy_t;

typedef void (* statefunc) (void *);

typedef struct statestruct
{
    boolean rotate;
    short   shapenum;           // a shapenum of -1 means get from ob->temp1
    short   tictime;
    void    (*think) (void *),(*action) (void *);
    struct  statestruct *next;
} statetype;


//---------------------
//
// trivial actor structure
//
//---------------------

typedef struct statstruct
{
    byte      tilex,tiley;
    short     shapenum;           // if shapenum == -1 the obj has been removed
    byte      *visspot;
    uint32_t  flags;
    byte      itemnumber;
#ifdef DARKANIMATION
    byte anim;
#endif
#ifdef KEYSFLOAT
    boolean   direct;
    int       position;
#endif
} statobj_t;


//---------------------
//
// door actor structure
//
//---------------------

typedef enum
{
    dr_open,dr_closed,dr_opening,dr_closing
} doortype;

typedef struct doorstruct
{
    byte     tilex,tiley;
    boolean  vertical;
    byte     lock;
    doortype action;
#ifdef BLAKESTYLEDOOR
    boolean doubledoor; 
    word     lposition;
#endif
    word     position;
    short    ticcount;
} doorobj_t;

typedef struct pushwallstruct
{
    word     state;      
    word     pos;        // amount a pushable wall has been moved (0-63)
    word     x;
    word     y;
    byte     dir;
    byte     tile;
#ifdef WALLPATCHES
    byte     patch;
#endif
} pwall_t;


//--------------------
//
// thinking actor structure
//
//--------------------

typedef struct objstruct
{
    activetype        active;
    short             ticcount;
    classtype         obclass;
    statetype         *state;
    uint32_t          flags;         // FL_SHOOTABLE, etc
    int32_t           distance;      // if negative, wait for that door to open
    dirtype           dir;
    fixed             x,y;
    word              tilex,tiley;
    byte              areanumber;
    short             viewx;
    word              viewheight;
    fixed             transx,transy; // in global coord
    short             angle;
    short             hitpoints;
    int32_t           speed;
    short             temp1,temp2,hidden;
    struct objstruct  *next,*prev;
} objtype;

enum
{
    bt_nobutton=-1,
    bt_attack=0,
    bt_strafe,
    bt_run,
    bt_use,
    bt_readyknife,
    bt_readypistol,
    bt_readymachinegun,
    bt_readychaingun,
    bt_readyrocketgun,
    bt_nextweapon,
    bt_prevweapon,
    bt_esc,
    bt_pause,
    bt_strafeleft,
    bt_straferight,
    bt_moveforward,
    bt_movebackward,
    bt_turnleft,
    bt_turnright,
    NUMBUTTONS
};


typedef enum
{
    wp_knife,
    wp_pistol,
    wp_machinegun,
    wp_chaingun,
    wp_rocketgun,
    NUMWEAPONS
} weapontype;


enum
{
    gd_baby,
    gd_easy,
    gd_medium,
    gd_hard,
    NUMDIFFICULTY
};

//---------------
//
// gamestate structure
//
//---------------

typedef struct
{
    short       difficulty, mapon, backto;
    int32_t     oldscore,score,nextextra;
    short       lives;
    short       health;
    short       ammo;
    short       ammo2;
    short       rockets;
    short       keys;
    short       medical;
    short       teleportnum;
    int         bjthrust,bjstrafe;
    boolean     outside;
#ifdef MAPCONTROLPARTIME 
    float       partime; 
#endif    
    weapontype  bestweapon,weapon,chosenweapon;
#ifdef SEAMLESS
    short       nextsecret;
#endif

    boolean     antidote;
    boolean     infected;
    boolean     bombdetector;

#ifdef FFDATAFIX
    int         ffTL, ffTR, ffBL, ffBR;
#endif

    short       faceframe,attackframe,attackcount,weaponframe;
    short       episode,secretcount,treasurecount,killcount,
                secrettotal,treasuretotal,killtotal;
    int32_t     TimeCount;
    int32_t     killx,killy;
    boolean     victoryflag;            // set during victory animations
    char        message[40];
} gametype;


typedef enum
{
    ex_stillplaying,
    ex_completed,
    ex_died,
    ex_warped,
    ex_resetgame,
    ex_loadedgame,
    ex_victorious,
    ex_abort,
    ex_demodone,
    ex_secretlevel
} exit_t;


extern word *mapsegs[MAPPLANES];
extern int mapon;

/*
=============================================================================

                             WL_MAIN DEFINITIONS

=============================================================================
*/

extern  boolean  loadedgame;
extern  fixed    focallength;
extern  int      viewscreenx, viewscreeny;
extern  int      viewwidth;
extern  int      viewheight;
extern  short    centerx;
extern  int32_t  heightnumerator;
extern  fixed    scale;

extern  int             AdjustX;  // To adjust images to the center,
extern  int             AdjustY;  // care of mathematical calculations

extern  int      dirangle[9];

extern  int      mouseadjustment;
extern  int      shootdelta;
extern  unsigned screenofs;

extern  boolean  startgame;
extern  char     str[80];
extern  char     configname[13];

//
// Command line parameter variables
//
extern  boolean  param_debugmode;
extern  boolean  param_nowait;
extern  int      param_difficulty;
extern  int      param_tedlevel;
extern  int      param_joystickindex;
extern  int      param_joystickhat;
extern  int      param_samplerate;
extern  int      param_audiobuffer;
extern  int      param_mission;
extern  boolean  param_goodtimes;
extern  boolean  param_ignorenumchunks;


void            NewGame (int difficulty,int episode);
void            CalcProjection (int32_t focal);
void            NewViewSize (int width);
void            CalculateAdjustXY (void);
boolean         SetViewSize (unsigned width, unsigned height);
boolean         LoadTheGame(FILE *file,int x,int y);
boolean         SaveTheGame(FILE *file,int x,int y);
void            ShowViewSize (int width);
void            ShutdownId (void);


/*
=============================================================================

                         WL_GAME DEFINITIONS

=============================================================================
*/

extern  gametype        gamestate;
extern  byte            bordercol;
extern  SDL_Surface     *latchpics[NUMLATCHPICS];
extern  char            demoname[13];
extern  char            hintname[13];
extern  char            dx8dir[9];
extern  char            dy8dir[9];
extern  char            dx4dir[4];
extern  char            dy4dir[4];

void    SetupGameLevel (void);
void    GameLoop (void);
void    DrawPlayBorder (void);
void    DrawStatusBorder (byte color);
void    DrawPlayScreen (void);
void    DrawPlayBorderSides (void);
void    ShowActStatus();
void    Set_Gamestate (void);
void    PlayDemo (int demonumber);
void    RecordDemo (void);
boolean InFloorRange (int value);

#ifdef SPEAR
extern  int32_t         spearx,speary;
extern  unsigned        spearangle;
extern  boolean         spearflag;
#endif

extern  boolean         NoShadingStuff;

#define ClearMemory SD_StopDigitized


// JAB
#define PlaySoundLocTile(s,tx,ty)       PlaySoundLocGlobal(s,(((int32_t)(tx) << TILESHIFT) + (1L << (TILESHIFT - 1))),(((int32_t)ty << TILESHIFT) + (1L << (TILESHIFT - 1))))
#define PlaySoundLocActor(s,ob)         PlaySoundLocGlobal(s,(ob)->x,(ob)->y)
void    PlaySoundLocGlobal(word s,fixed gx,fixed gy);
void    UpdateSoundLoc(void);

/*
=============================================================================

                            WL_PLAY DEFINITIONS

=============================================================================
*/

extern  byte            tilemap[MAPSIZE][MAPSIZE];      // wall values only
extern  byte            spotvis[MAPSIZE][MAPSIZE];
extern  objtype         *actorat[MAPSIZE][MAPSIZE];
extern  objtype         *player;
extern  unsigned        tics;
extern  int             viewsize;
extern  int             lastgamemusicoffset;

extern int TeleportX1A,TeleportY1A,TeleportX1B,TeleportY1B;
extern int TeleportX2A,TeleportY2A,TeleportX2B,TeleportY2B;
extern int TeleportX3A,TeleportY3A,TeleportX3B,TeleportY3B;
extern int TeleportX4A,TeleportY4A,TeleportX4B,TeleportY4B;
extern int TeleLaddX1A,TeleLaddY1A,TeleLaddX1B,TeleLaddY1B;
extern int TeleLaddX2A,TeleLaddY2A,TeleLaddX2B,TeleLaddY2B;

//
// current user input
//
extern  int         controlx,controly;              // range from -100 to 100
extern  boolean     buttonstate[NUMBUTTONS];
extern  objtype     objlist[MAXACTORS];
extern  boolean     buttonheld[NUMBUTTONS];
extern  exit_t      playstate;
extern  boolean     madenoise;
extern  statobj_t   statobjlist[MAXSTATS];
extern  statobj_t   *laststatobj;
extern  objtype     *newobj,*killerobj;
extern  doorobj_t   doorobjlist[MAXDOORS];
extern  doorobj_t   *lastdoorobj;
extern  int         godmode;
extern  boolean     demorecord,demoplayback;
extern  int8_t      *demoptr, *lastdemoptr;
extern  memptr      demobuffer;

//
// control info
//
extern  boolean     mouseenabled,joystickenabled;
extern  int         dirscan[4];
extern  int         buttonscan[NUMBUTTONS];
extern  int         buttonmouse[4];
extern  int         buttonjoy[32];

void    InitActorList (void);
void    GetNewActor (void);
void    PlayLoop (void);

void    CenterWindow(word w,word h);

void    InitRedShifts (void);
void    FinishPaletteShifts (void);

void    RemoveObj (objtype *gone);
void    PollControls (void);
int     StopMusic(void);
void    StartMusic(void);
void    ContinueMusic(int offs);
void    StartDamageFlash (int damage);
void    StartBonusFlash (void);

#ifdef SPEAR
extern  int32_t     funnyticount;           // FOR FUNNY BJ FACE
#endif
extern  objtype     *objfreelist;     // *obj,*player,*lastobj,
extern  boolean     noclip,ammocheat;
extern  int         singlestep, extravbls;


#ifdef TEKZOOMSHOCK  // BEGIN ZOOM BLOCK 
// Global functions 
void   BasicZoom( int power );          // Sets the basic zoom (power: 0..100) 
void   StartDamageShock( int power );   // Starts the shock effect (power: 0..300) 
void   StartRecoilEffect( int power );  // Starts the recoil effect (power: 0..100) 
void   StartQuakeEffect( int power );   // Starts the quake effect (power: 0..200) 
void   ResetZoomEffects( void );        // Resets all the zoom effects immediately 
void   UpdateZoomEffects( void );       // Updates all the zoom effects in one pass 
int      GetShockedAngle( void );       // Gets the viewing angle, including shock and quake effects 
#endif  // END ZOOM BLOCK

/*
=============================================================================

                                WL_INTER

=============================================================================
*/

void IntroScreen (void);
void PG13(void);
void DrawHighScores(void);
void CheckHighScore (int32_t score,word other);
void Victory (void);
void LevelCompleted (void);
void ClearSplitVWB (void);

void PreloadGraphics(void);


/*
=============================================================================

                                WL_DEBUG

=============================================================================
*/

int DebugKeys (void);
void GetCurrentRatio (void);

/*
=============================================================================

                            WL_DRAW DEFINITIONS

=============================================================================
*/
// math tables
extern  short     *pixelangle;
extern  int32_t   finetangent[FINEANGLES/4];
extern  fixed     sintable[];
extern  fixed     *costable;
extern  int       *wallheight;
extern  word      horizwall[],vertwall[];
extern  int32_t   lasttimecount;
extern  int32_t   frameon;

extern  unsigned  screenloc[3];

extern  boolean   fizzlein, fpscounter;
extern  fixed     viewx,viewy;                    // the focal point
extern  fixed     viewsin,viewcos;
extern  byte      door_tex[dr_types];
extern  byte      door_sides[dr_types];
extern  byte      door_type[dr_types];
extern  boolean   leftside;
#ifdef CHRISAUTOMAPTEK
extern  boolean   WalkedFloor[NUMAREAS];
#endif
#ifdef CHRISAUTOMAP2
extern  boolean   BeenThere [MAPSIZE][MAPSIZE];
#endif 

void    ThreeDRefresh (void);
void    CalcTics (void);

void    DrawCrossHair (void);
void    GetMessage (char *lastmessage);
void    DrawMessage (void);

#ifdef FIX3D
typedef struct
{
    word leftpix,rightpix;
    word dataofs[64];
// table data after dataofs[rightpix-leftpix+1]
} t_compshape;
#endif

void    LightTableSetup(void);
void    SetLeftDoorSide (void); 
void    SetRightDoorSide (void);
void    DrawLevelMap(void);
/*
=============================================================================

                             WL_STATE DEFINITIONS

=============================================================================
*/
void    InitHitRect (objtype *ob, unsigned radius);
void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
void    SelectChaseDir (objtype *ob);
void    SelectDodgeDir (objtype *ob);
void    SelectRunDir (objtype *ob);
void    MoveRatObj (objtype *ob, int32_t move);
void    MoveObj (objtype *ob, int32_t move);
boolean SightPlayer (objtype *ob);

void    KillActor (objtype *ob);
void    DamageActor (objtype *ob, unsigned damage);

boolean CheckLine (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                             WL_AGENT DEFINITIONS

=============================================================================
*/

extern  short    anglefrac;
extern  int      facecount, facetimes;
extern  word     plux,pluy;         // player coordinates scaled to unsigned
extern  int32_t  thrustspeed;
extern  objtype  *LastAttacker;

void    Thrust (int angle, int32_t speed);
void    SpawnPlayer (int tilex, int tiley, int dir);
void    TakeDamage (int points,objtype *attacker);
void    GivePoints (int32_t points);
void    GetBonus (statobj_t *check);
void    GiveWeapon (int weapon);
void    GiveAmmo (int ammo);
void    GiveKey (int key);
static  fixed FixedByFracOrig(fixed a, fixed b);
//
// player state info
//

void    StatusDrawFace(unsigned picnum);
void    DrawFace (void);
void    DrawHealth (void);
void    HealSelf (int points);
void    DrawLevel (void);
void    DrawLives (void);
void    GiveExtraMan (void);
void    DrawScore (void);
void    DrawWeapon (void);
void    DrawKeys (void);
void    DrawAmmo (void);


/*
=============================================================================

                             WL_ACT1 DEFINITIONS

=============================================================================
*/

extern  doorobj_t doorobjlist[MAXDOORS];
extern  doorobj_t *lastdoorobj;
extern  short     doornum;
extern  pwall_t   pwall;
extern  byte      areaconnect[NUMAREAS][NUMAREAS];
extern  boolean   areabyplayer[NUMAREAS];

void InitDoorList (void);
void InitStaticList (void);
void SpawnStatic (int tilex, int tiley, int type);
void SpawnDoor (int tilex, int tiley, boolean vertical, int lock);
void MoveDoors (void);
void MovePWalls (void);
void OpenDoor (int door);
void PlaceItemType (int itemtype, int tilex, int tiley);
void PushWall (int checkx, int checky, int dir);
void OperateDoor (int door);
void InitAreas (void);

/*
=============================================================================

                             WL_ACT2 DEFINITIONS

=============================================================================
*/
extern  statetype s_blood0;
extern  statetype s_grddie1;
extern  statetype s_grddie4;
extern  statetype s_madmeddie1;
extern  statetype s_madmedblocker;
extern  statetype s_madmedstand;
#ifdef DARREN
extern  statetype s_grd2die1;
#endif
extern  statetype s_dogdie1;
extern  statetype s_dogdead;
extern  statetype s_ofcdie1;
extern  statetype s_ofcdie5;
extern  statetype s_mutdie1;
extern  statetype s_mutdie5;
extern  statetype s_ssdie1;
extern  statetype s_ssdie4;
extern  statetype s_bossdie1;
extern  statetype s_schabbdie1;
extern  statetype s_fakedie1;
extern  statetype s_mechadie1;
extern  statetype s_hitlerdie1;
extern  statetype s_greteldie1;
extern  statetype s_giftdie1;
extern  statetype s_fatdie1;

extern  statetype s_spectredie1;
extern  statetype s_angeldie1;
extern  statetype s_transdie0;
extern  statetype s_uberdie0;
extern  statetype s_willdie1;
extern  statetype s_deathdie1;

extern  statetype s_gatlerdie1;
extern  statetype s_barrel1die1;
extern  statetype s_barrel2die1;
extern  statetype s_plantdie0;

extern  statetype s_grdchase1;
extern  statetype s_madmedrise1;
extern  statetype s_madmedchase1;
#ifdef DARREN
extern  statetype s_grd2chase1;
#endif
extern  statetype s_dogchase1;
extern  statetype s_ofcchase1;
extern  statetype s_sschase1;
extern  statetype s_mutchase1;
extern  statetype s_bosschase1;
extern  statetype s_schabbchase1;
extern  statetype s_fakechase1;
extern  statetype s_mechachase1;
extern  statetype s_gretelchase1;
extern  statetype s_giftchase1;
extern  statetype s_fatchase1;

extern  statetype s_spectrechase1;
extern  statetype s_angelchase1;
extern  statetype s_transchase1;
extern  statetype s_uberchase1;
extern  statetype s_willchase1;
extern  statetype s_deathchase1;

extern  statetype s_blinkychase1;
extern  statetype s_pinkychase1;
extern  statetype s_inkychase1;
extern  statetype s_clydechase1;
extern  statetype s_dogjump1;

extern  statetype s_hitlerchase1;

extern  statetype s_gatlerchase1;

extern  statetype s_grdpain;
extern  statetype s_grdpain1;
#ifdef DARREN
extern  statetype s_grd2pain;
extern  statetype s_grd2pain1;
#endif
extern  statetype s_ofcpain;
extern  statetype s_ofcpain1;
extern  statetype s_sspain;
extern  statetype s_sspain1;
extern  statetype s_mutpain;
extern  statetype s_mutpain1;

extern  statetype s_deathcam;

extern  statetype s_schabbdeathcam2;
extern  statetype s_hitlerdeathcam2;
extern  statetype s_giftdeathcam2;
extern  statetype s_fatdeathcam2;

extern  statetype s_jawsopen;
extern  statetype s_jawsclosed;
extern  statetype s_jaws2open;
extern  statetype s_jaws2closed;

void SpawnStand (enemy_t which, int tilex, int tiley, int dir);
void SpawnPatrol (enemy_t which, int tilex, int tiley, int dir);
#ifdef USERATS
void SpawnAmbPatrol (enemy_t which, int tilex, int tiley, int dir);
#endif
void KillActor (objtype *ob);

void SpawnDeadGuard (int tilex, int tiley, statetype *state);
void SpawnJaws (int jawstype, int tilex, int tiley);
void SpawnMadMed (int tilex, int tiley);
void SpawnDarren (int tilex, int tiley);
void SpawnBoss (int tilex, int tiley);
void SpawnGretel (int tilex, int tiley);
void SpawnTrans (int tilex, int tiley);
void SpawnUber (int tilex, int tiley);
void SpawnWill (int tilex, int tiley);
void SpawnDeath (int tilex, int tiley);
void SpawnAngel (int tilex, int tiley);
void SpawnSpectre (int tilex, int tiley);
void SpawnGhosts (int which, int tilex, int tiley);
void SpawnSchabbs (int tilex, int tiley);
void SpawnGift (int tilex, int tiley);
void SpawnFat (int tilex, int tiley);
void SpawnFakeHitler (int tilex, int tiley);
void SpawnHitler (int tilex, int tiley);
void SpawnNormalHitler (int tilex, int tiley);

void SpawnGatler (int tilex, int tiley);
void SpawnBarrel1 (int tilex, int tiley);
void SpawnBarrel2 (int tilex, int tiley);
void SpawnPlant1 (int tilex, int tiley);
void SpawnPlant2 (int tilex, int tiley);
void SpawnPlant3 (int tilex, int tiley);
void SpawnPlant4 (int tilex, int tiley);

void A_DeathScream (objtype *ob);
void SpawnBJVictory (void);

/*
=============================================================================

                             WL_TEXT DEFINITIONS

=============================================================================
*/

extern  char    helpfilename[],endfilename[],hintname[];
extern  void    HelpScreens(void);
extern  void    EndText(void);


/*
=============================================================================

                            DREAMCAST DEFINITIONS

=============================================================================
*/

#ifdef _arch_dreamcast

// defined in dc_main.cpp
void DC_Main(void);
void DC_CheckParameters(void);
int DC_MousePresent(void);

// defined in dc_vmu.cpp
void StatusDrawLCD(int lcd_index);
int DC_SaveToVMU(char *src, int tp);
int DC_LoadFromVMU(char *dst);

#else

// ignore calls to this function
#define StatusDrawLCD(x)

#endif


/*
=============================================================================

                               GP2X DEFINITIONS

=============================================================================
*/

#ifdef GP2X

void GP2X_Init();
void GP2X_Shutdown();
void GP2X_ButtonDown(int button);
void GP2X_ButtonUp(int button);

#endif


/*
=============================================================================

                             MISC DEFINITIONS

=============================================================================
*/

static inline fixed FixedMul(fixed a, fixed b)
{
	return (fixed)(((int64_t)a * b + 0x8000) >> 16);
}

#ifdef PLAYDEMOLIKEORIGINAL
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) ((demorecord || demoplayback) ? (orig) : (sdl))
    #define DEMOCOND_ORIG                  (demorecord || demoplayback)
    #define DEMOIF_SDL                     if(DEMOCOND_SDL)
#else
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) (sdl)
    #define DEMOCOND_ORIG                  false
    #define DEMOIF_SDL
#endif
#define DEMOCOND_SDL                   (!DEMOCOND_ORIG)

#define GetTicks() ((SDL_GetTicks()*7)/100)

#define ISPOINTER(x) ((((uintptr_t)(x)) & ~0xffff) != 0)

#define CHECKMALLOCRESULT(x) if(!(x)) Quit("Out of memory at %s:%i", __FILE__, __LINE__)

#ifdef _WIN32
    #define strcasecmp stricmp
    #define strncasecmp strnicmp
#else
    static inline char* itoa(int value, char* string, int radix)
    {
	    sprintf(string, "%d", value);
	    return string;
    }

    static inline char* ltoa(long value, char* string, int radix)
    {
	    sprintf(string, "%ld", value);
	    return string;
    }
#endif

#define lengthof(x) (sizeof(x) / sizeof(*(x)))
#define endof(x)    ((x) + lengthof(x))

static inline word READWORD(byte *&ptr)
{
    word val = ptr[0] | ptr[1] << 8;
    ptr += 2;
    return val;
}

static inline longword READLONGWORD(byte *&ptr)
{
    longword val = ptr[0] | ptr[1] << 8 | ptr[2] << 16 | ptr[3] << 24;
    ptr += 4;
    return val;
}


/*
=============================================================================

                           FEATURE DEFINITIONS

=============================================================================
*/

#ifdef USE_FEATUREFLAGS
    // The currently available feature flags
    #define FF_STARSKY      0x0001
    #define FF_PARALLAXSKY  0x0002
    #define FF_CLOUDSKY     0x0004
    #define FF_RAIN         0x0010
    #define FF_SNOW         0x0020

    // The ffData... variables contain the 16-bit values of the according corners of the current level.
    // The corners are overwritten with adjacent tiles after initialization in SetupGameLevel
    // to avoid interpretation as e.g. doors.
    extern int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight;

    /*************************************************************
     * Current usage of ffData... variables:
     * ffDataTopLeft:     lower 8-bit: ShadeDefID
     * ffDataTopRight:    FeatureFlags
     * ffDataBottomLeft:  CloudSkyDefID or ParallaxStartTexture
     * ffDataBottomRight: unused
     *************************************************************/

    // The feature flags are stored as a wall in the upper right corner of each level
    static inline word GetFeatureFlags()
    {
#ifdef FFTWOPARTS
        return ((ffDataTopRight) & 0x00ff);
#else
        return ffDataTopRight;
#endif
    }

#endif

#ifdef USE_FLOORCEILINGTEX
    void DrawFloorAndCeiling(byte *vbuf, unsigned vbufPitch, int min_wallheight);
#endif

#ifdef USE_PARALLAX
    void DrawParallax(byte *vbuf, unsigned vbufPitch);
#endif

#ifdef USE_DIR3DSPR
    void Scale3DShape(byte *vbuf, unsigned vbufPitch, statobj_t *ob);
#ifdef DIRECTIONALACTORS
    void Scale3DShapeActor(byte *vbuf, unsigned vbufPitch, objtype *ob, short ThreeDimFlag);
#endif
#endif

#endif
