// WL_DRAW.CPP

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_atmos.h"
#include "wl_shade.h"

/**************************************************************************
**                          LOCAL CONSTANTS                              **
**************************************************************************/ 

void TransformActor (objtype *ob);
void BuildTables (void);
void ClearScreen (void);
int  CalcRotate (objtype *ob);
void DrawScaleds (void);
void CalcTics (void);
void ThreeDRefresh (void);

int  weaponscale[NUMWEAPONS] = {SPR_KNIFEREADY, SPR_PISTOLREADY,
                                SPR_MACHINEGUNREADY, SPR_CHAINREADY, SPR_ROCKREADY};


#define DOOR_TYPES    (dr_types)

byte door_tex[DOOR_TYPES]=
{
  DRNORM1DOORPIC, DRLOCK1DOORPIC, DRLOCK2DOORPIC, DRLOCK3DOORPIC,
  DRLOCK4DOORPIC, DRELEVDOORPIC,  DRNORM2DOORPIC, DRNORM3DOORPIC
};
byte door_sides[DOOR_TYPES]=
{
  DRNORM1SIDEPIC, DRLOCK1SIDEPIC, DRLOCK2SIDEPIC, DRLOCK3SIDEPIC,
  DRLOCK4SIDEPIC, DRELEVSIDEPIC,  DRNORM2SIDEPIC, DRNORM3SIDEPIC
};
byte door_type[DOOR_TYPES]=
{
  DRNORM1TYPE, DRLOCK1TYPE, DRLOCK2TYPE, DRLOCK3TYPE,
  DRLOCK4TYPE, DRELEVTYPE,  DRNORM2TYPE, DRNORM3TYPE
};
/**************************************************************************
**                          GLOBAL VARIABLES                             **
**************************************************************************/ 

static byte *vbuf = NULL;
unsigned    vbufPitch = 0;
int32_t     lasttimecount;
int32_t     frameon;
int         messagetime=0;
int         TextXVar, TextYVar;
boolean     fpscounter;
int         fps_frames=0, 
            fps_time=0, 
            fps=0;
int         *wallheight;
int         min_wallheight;
//
// math tables
//
short       *pixelangle;
int32_t     finetangent[FINEANGLES/4];
fixed       sintable[ANGLES+ANGLES/4];
fixed       *costable = sintable+(ANGLES/4);
//
// refresh variables
//
fixed       viewx,viewy;                    // the focal point
short       viewangle;
fixed       viewsin,viewcos;
//
// wall optimization variables
//
int         lastside;               // true for vertical
int32_t     lastintercept;
int         lasttilehit;
int         lasttexture;
byte        *postsource;
int         postx;
int         postwidth;
//
// ray tracing variables
//
short       focaltx,focalty,viewtx,viewty;
longword    xpartialup,xpartialdown,ypartialup,ypartialdown;
short       midangle,angle;
word        tilehit;
int         pixx;
short       xtile,ytile;
short       xtilestep,ytilestep;
int32_t     xintercept,yintercept;
word        xstep,ystep;
word        xspot,yspot;
int         texdelta;

word horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];

#ifdef TEKZOOMSHOCK   // BEGIN ZOOM BLOCK 
// Global definitions 
#ifndef     FOCALLENGTH 
  #define   FOCALLENGTH            0x5700L               // MUST match the FOCALLENGTH definition in WL_MAIN.C 
#endif 
#define     FOCALTHRESHOLD         FOCALLENGTH            // Greater lengths won't change the view position 
#endif // END ZOOM BLOCK 

#ifndef FIX3D
typedef struct
{
  word      leftpix,rightpix;
  word      dataofs[64]; // table data after dataofs[rightpix-leftpix+1]
} t_compshape;
#endif

typedef struct
{
  short     viewx,
            viewheight,
            shapenum,
            flags;          // this must be changed to uint32_t, when you
                            // you need more than 16-flags for drawing
#ifdef USE_DIR3DSPR
  statobj_t *transsprite;
  objtype   *transactor;
#endif
} visobj_t;

visobj_t    vislist[MAXVISABLE];
visobj_t    *visptr,*visstep,*farthest;

/**************************************************************************
**                    3 - D  D E F I N I T I O N S                       **
**************************************************************************/ 



/**************************************************************************
**                         Math Functions                                **
**************************************************************************/ 

/* ======================================================================
= TransformActor
=
= Takes paramaters:
=   gx,gy               : globalx/globaly of point
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================================================================= */
void TransformActor (objtype *ob)
{
  fixed nx,ny;

// translate point to view centered coordinates
//  ob->x-viewx;
//  ob->y-viewy;
// calculate newx
  nx = FixedMul(ob->x-viewx,viewcos)-FixedMul(ob->y-viewy,viewsin)-ACTORSIZE;         
  // - ACTORSIZE = fudge the shape forward a bit, because the 
  // midpoint could put parts of the shape into an adjacent wall
// calculate newy
  ny = FixedMul(ob->y-viewy,viewcos)+FixedMul(ob->x-viewx,viewsin);
// calculate perspective ratio
  ob->transx = nx;
  ob->transy = ny;
  if (nx<MINDIST)                 // too close, don't overflow the divide
  {
    ob->viewheight = 0;
    return;
  }
  ob->viewx = (word)(centerx + ny*scale/nx);
// calculate height (heightnumerator/(nx>>8))
  ob->viewheight = (word)(heightnumerator/(nx>>8));
}

/* ======================================================================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty               : tile the object is centered in
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================================================================= */
boolean TransformTile (int tx, int ty, short *dispx, short *dispheight)
{
  fixed nx,ny;
//
// translate point to view centered coordinates
//
//  x = ((int32_t)tx<<TILESHIFT)+0x8000-viewx;
//  y = ((int32_t)ty<<TILESHIFT)+0x8000-viewy;
// calculate newx
  nx = FixedMul(((int32_t)tx<<TILESHIFT)+0x8000-viewx,viewcos) -
       FixedMul(((int32_t)ty<<TILESHIFT)+0x8000-viewy,viewsin) -0x2000; 
       // 0x2000 is size of object
// calculate newy
  ny = FixedMul(((int32_t)ty<<TILESHIFT)+0x8000-viewy,viewcos) +
       FixedMul(((int32_t)tx<<TILESHIFT)+0x8000-viewx,viewsin);
// calculate height / perspective ratio
  if (nx<MINDIST) *dispheight = 0; // too close, don't overflow the divide
  else
  {
    *dispx = (short)(centerx + ny*scale/nx);
    *dispheight = (short)(heightnumerator/(nx>>8));
  }
// see if it should be grabbed
  if (nx<TILEGLOBAL && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2) return true;
  else return false;
}

/* ======================================================================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
=========================================================================*/
int CalcHeight()
{
    fixed z = FixedMul(xintercept - viewx, viewcos) - 
              FixedMul(yintercept - viewy, viewsin);
    if(z < MINDIST) z = MINDIST;
    int height = heightnumerator / (z >> 8);
    if(height < min_wallheight) min_wallheight = height;
    return height;
}

/* 
=====================
=
= CalcRotate
=
=====================
*/
int CalcRotate (objtype *ob)
{
  int angle, viewangle;
  // this isn't exactly correct, as it should vary by a trig value,
  // but it is close enough with only eight rotations

  viewangle = player->angle + (centerx - ob->viewx)/8;

  if (ob->obclass == rocketobj || ob->obclass == hrocketobj)
    angle = (viewangle-180) - ob->angle;
  else
    angle = (viewangle-180) - dirangle[ob->dir];

  angle+=ANGLES/16;
  while (angle>=ANGLES) angle-=ANGLES;
  while (angle<0) angle+=ANGLES;
// 2 rotation pain frame pain with shooting frame bugfix
  if (ob->state->rotate == 2) return 0;            
  return angle/(ANGLES/8);
}

/*
=====================
=
= CalcTics
=
=====================
*/
void CalcTics (void)
{
// calculate tics since last refresh for adaptive timing
  // if the game was paused a LONG time
  if (lasttimecount > (int32_t) GetTimeCount())
      lasttimecount = GetTimeCount();    

  uint32_t curtime = SDL_GetTicks();
  tics = (curtime * 7) / 100 - lasttimecount;
  if(!tics)
  { // wait until end of current tic
    SDL_Delay(((lasttimecount + 1) * 100) / 7 - curtime);
    tics = 1;
  }
  lasttimecount += tics;
  if (tics>MAXTICS) tics = MAXTICS;
}

/*************************************************************************/

/**************************************************************************
**                      VGA CLEARING THE SCREEN                          **
**************************************************************************/ 
byte ceiling[]=
{
#ifndef SPEAR
#ifdef WOLFDX
 0x0a,0x0a,0x0a,0x0a,0x0a,0x0a,0x0a,0x0a,0x0a,0xbf,
 0x4e,0x4e,0x4e,0x0a,0x8d,0x4e,0x0a,0x2d,0x0a,0x8d,
 0x0a,0x0a,0x0a,0x0a,0x0a,0x2d,0xdd,0x0a,0x0a,0x98,

 0x0a,0x9d,0x2d,0xdd,0xdd,0x9d,0x2d,0x4d,0x0a,0xdd,
 0x7d,0x0a,0x2d,0x2d,0xdd,0xd7,0x0a,0x0a,0x0a,0x2d,
 0x0a,0x0a,0x0a,0x0a,0xdd,0xdd,0x7d,0xdd,0xdd,0xdd
#else
 0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0xbf,
 0x4e,0x4e,0x4e,0x1d,0x8d,0x4e,0x1d,0x2d,0x1d,0x8d,
 0x1d,0x1d,0x1d,0x1d,0x1d,0x2d,0xdd,0x1d,0x1d,0x98,

 0x1d,0x9d,0x2d,0xdd,0xdd,0x9d,0x2d,0x4d,0x1d,0xdd,
 0x7d,0x1d,0x2d,0x2d,0xdd,0xd7,0x1d,0x1d,0x1d,0x2d,
 0x1d,0x1d,0x1d,0x1d,0xdd,0xdd,0x7d,0xdd,0xdd,0xdd
#endif
#else
 0x6f,0x4f,0x1d,0xde,0xdf,0x2e,0x7f,0x9e,0xae,0x7f,
 0x1d,0xde,0xdf,0xde,0xdf,0xde,0xe1,0xdc,0x2e,0x1d,0xdc
#endif
};
/* ======================= VGAClearScreen ====================== */
void VGAClearScreen (void)
{
  int y;
  byte *ptr = vbuf;
#ifdef USE_SHADING
  for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
    memset(ptr, shadetable[GetShade((viewheight / 2 - y) << 3)][ceiling[gamestate.episode*10+mapon]], viewwidth);
  for(; y < viewheight; y++, ptr += vbufPitch)
#ifdef WOLFDX
        memset(ptr, shadetable[GetShade((y - viewheight / 2) << 3)][0x0c], viewwidth);
#else
        memset(ptr, shadetable[GetShade((y - viewheight / 2) << 3)][0x19], viewwidth);
#endif
#else
  for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
    memset(ptr, ceiling[gamestate.episode*10+mapon], viewwidth);
  for(; y < viewheight; y++, ptr += vbufPitch)
#ifdef WOLFDX
        memset(ptr, 0x0c, viewwidth);
#else
        memset(ptr, 0x19, viewwidth);
#endif
#endif
}

/**************************************************************************
**                   Wall Drawing & Scaling Routines                     **
**************************************************************************/ 
void ScalePost()
{
    int ywcount, yoffs, yw, yd, yendoffs;
    byte col;

#ifdef USE_SHADING
    byte *curshades = shadetable[GetShade(wallheight[postx])];
#endif

    ywcount = yd = wallheight[postx] >> 3;
    if(yd <= 0) yd = 100;

    yoffs = (viewheight / 2 - ywcount) * vbufPitch;
    if(yoffs < 0) yoffs = 0;
    yoffs += postx;

    yendoffs = viewheight / 2 + ywcount - 1;
    yw=TEXTURESIZE-1;

    while(yendoffs >= viewheight)
    {
        ywcount -= TEXTURESIZE/2;
        while(ywcount <= 0)
        {
            ywcount += yd;
            yw--;
        }
        yendoffs--;
    }
    if(yw < 0) return;

#ifdef USE_SHADING
    col = curshades[postsource[yw]];
#else
    col = postsource[yw];
#endif
    yendoffs = yendoffs * vbufPitch + postx;
    while(yoffs <= yendoffs)
    {
        vbuf[yendoffs] = col;
        ywcount -= TEXTURESIZE/2;
        if(ywcount <= 0)
        {
            do
            {
                ywcount += yd;
                yw--;
            }
            while(ywcount <= 0);
            if(yw < 0) break;
#ifdef USE_SHADING
            col = curshades[postsource[yw]];
#else
            col = postsource[yw];
#endif
        }
        yendoffs -= vbufPitch;
    }
}

void GlobalScalePost(byte *vidbuf, unsigned pitch)
{
  vbuf = vidbuf;
  vbufPitch = pitch;
  ScalePost();
}

/**************************************************************************
**                        Scale SPRITE Shapes                            **
**************************************************************************/ 
void ScaleShape (int xcenter, int shapenum, unsigned height, uint32_t flags)
{
    t_compshape *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    byte *vmem;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;

#ifdef USE_SHADING
    byte *curshades;
    if(flags & FL_FULLBRIGHT)
        curshades = shadetable[0];
    else
        curshades = shadetable[GetShade(height)];
#endif

    shape = (t_compshape *) PM_GetSprite(shapenum);

    scale=height>>3;                 // low three bits are fractional
    if(!scale) return;   // too close or far away

    pixheight=scale*SPRITESCALEFACTOR;
    actx=xcenter-scale;
    upperedge=viewheight/2-scale;

    cmdptr=(word *) shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline=(byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                if(wallheight[lpix]<=(int)height)
                {
                    line=cline;
                    while((endy = READWORD(line)) != 0)
                    {
                        endy >>= 1;
                        newstart = READWORD(line);
                        starty = READWORD(line) >> 1;
                        j=starty;
                        ycnt=j*pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(screndy<0) vmem=vbuf+lpix;
                        else vmem=vbuf+screndy*vbufPitch+lpix;
                        for(;j<endy;j++)
                        {
                            scrstarty=screndy;
                            ycnt+=pixheight;
                            screndy=(ycnt>>6)+upperedge;
                            if(scrstarty!=screndy && screndy>0)
                            {
#ifdef USE_SHADING
                                col=curshades[((byte *)shape)[newstart+j]];
#else
                                col=((byte *)shape)[newstart+j];
#endif
                                if(scrstarty<0) scrstarty=0;
                                if(screndy>viewheight) screndy=viewheight,j=endy;

                                while(scrstarty<screndy)
                                {
                                    *vmem=col;
                                    vmem+=vbufPitch;
                                    scrstarty++;
                                }
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

/**************************************************************************
**                 Simple Scaling for Drawing Weapons                    **
**************************************************************************/ 
void SimpleScaleShape (int xcenter, int shapenum, unsigned height)
{
    t_compshape   *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;
    byte *vmem;

    shape = (t_compshape *) PM_GetSprite(shapenum);

    scale=height>>1;
    pixheight=scale*SPRITESCALEFACTOR;
    actx=xcenter-scale;
    upperedge=viewheight/2-scale;

    cmdptr=shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline = (byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                line=cline;
                while((endy = READWORD(line)) != 0)
                {
                    endy >>= 1;
                    newstart = READWORD(line);
                    starty = READWORD(line) >> 1;
                    j=starty;
                    ycnt=j*pixheight;
                    screndy=(ycnt>>6)+upperedge;
                    if(screndy<0) vmem=vbuf+lpix;
                    else vmem=vbuf+screndy*vbufPitch+lpix;
                    for(;j<endy;j++)
                    {
                        scrstarty=screndy;
                        ycnt+=pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(scrstarty!=screndy && screndy>0)
                        {
                            col=((byte *)shape)[newstart+j];
                            if(scrstarty<0) scrstarty=0;
                            if(screndy>viewheight) screndy=viewheight,j=endy;

                            while(scrstarty<screndy)
                            {
                                *vmem=col;
                                vmem+=vbufPitch;
                                scrstarty++;
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

/*************************************************************************/

/**************************************************************************
**               Wall & Door Texture Selection & Scaling                 **
**************************************************************************/ 
// ------------------------------------------------------------------------
//                     64 Plus Wall values Routine 
// ------------------------------------------------------------------------
//
//  A means of tricking the system into allowing us access to bit 6
//  ie:  If the wall being checked is a Pushwall, bit's 6 & 7 will be 1
//       If the wall being checked is a Door, then Bit 7 will be 1
//  Chris took BrotherTank's original work and made it better.  
//  Here are those routines: 
//
//int CheckHighBits (int value)       // Check For Door Bit 7 only 
//{ 
//   if ((value & 0x40) && (value & 0x80)) // If Pushwall Return False Bit7&6 
//      { return 0; } 
//   else 
//     { 
//   if (value & 0x80)        // Is Door Return True 
//     { return 1; } 
//   else 
//     { return 0; }                  // Not Door Return False 
//      } 
//} 
//
//int CheckAdjacentTile (int x, int y) 
//{  // Part of 64+ Wall Tiles - Make sure within Map Boundaries 
//   // to fix bug of not allowing 64+ in outer edge of map 
//   // Checks Adjacent Tile for Bit 7 Set - Door or Pushwall 
//   if ((y-1 >= 0) && (tilemap[x][y-1] & 0x80) ||
//      (y+1 <= 63) && (tilemap[x][y+1] & 0x80) || 
//      (x-1 >= 0) && (tilemap[x-1][y] & 0x80) ||
//      (x+1 <= 63) && (tilemap[x+1][y] & 0x80)) 
//    return 1; 
//   else 
//    return 0; 
//}
//
//
byte CheckHighBits (byte highbit)  // Check For Door Bit 7 only
{
  if (highbit/64 == 2) return 1; else return 0; // True = Bit 7 Set  False = Bit Not Set
} 

boolean CheckVertTile (int x, int y) 
{ 
  if (((x > 0) && (tilemap[x-1][y]/64 == 2) && 
       (!doorobjlist[tilemap[x-1][y]-128].vertical)) ||
      ((x+1 < MAPSIZE) && (tilemap[x+1][y]/64 == 2) && 
       (!doorobjlist[tilemap[x+1][y]-128].vertical))) return 1; 
  return 0; 
}

boolean CheckHorizTile (int x, int y) 
{ 
  if (((y > 0) && (tilemap[x][y-1]/64 == 2) && 
       (doorobjlist[tilemap[x][y-1]-128].vertical)) ||
      ((y+1 < MAPSIZE) && (tilemap[x][y+1]/64 == 2) && 
       (doorobjlist[tilemap[x][y+1]-128].vertical))) return 1; 
  return 0; 
}

#ifdef WALLPATCHES
/**************************************************************************
**                Applying Wall Patches to Wall Images                   **
**************************************************************************/ 
byte     patchedwall[PMPageSize]; 
int      lastpatchnum; 
// =================== Get Patch Number ====================== 
int GetPatchNum(void) 
{ 
  int patch = MAPSPOT(xtile,ytile,1);
  if (patch >= PFIRSTPATCH && patch <= PLASTPATCH) return (patch - (PFIRSTPATCH - FIRSTPATCH));
  if (patch >= FIRSTPATCH && patch <= LASTPATCH && patch != LASTPATCH) return patch; 
  return 0; 
}
// =================== Apply the Patch ====================== 
boolean ApplyPatch(int wallpic, boolean vertical) 
{ 
  byte   *scan; 
  int    pixel, patch; 

  patch = GetPatchNum(); 
  if (patch > 0 && wallpic < DOORWALL) 
  { 
    lastpatchnum = patch; 
    patch = ((patch - FIRSTPATCH) * 2) + vertical; 
    scan = PM_GetTexture(wallpic); // Get the Wall 
    for (pixel=0; pixel<PMPageSize; pixel++,scan++) { patchedwall[pixel] = *scan; }
    scan = PM_GetTexture(PATCHWALL+patch); // Apply the Patch
    for (pixel=0; pixel<PMPageSize; pixel++,scan++)
    {
      if (*scan < 255) patchedwall[pixel] = *scan;
    }
    return true; 
  } 
  lastpatchnum = 0; 
  return false; 
}
#endif

/**************************************************************************
**                      Drawing the WALLS                                **
**************************************************************************/ 
void HitVertWall (void)
{
  int wallpic;
  int texture;

  ytile = (short) (yintercept >> TILESHIFT);
  texture = ((yintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

  if (xtilestep == -1)
  {
    if(MAPSPOT(xtile,ytile,1) != TILEMIRROR) texture = TEXTUREMASK-texture;
    xintercept += TILEGLOBAL;
  }

  if (lastside==1 && lastintercept==xtile && 
      lasttilehit==tilehit && !CheckVertTile(xtile,ytile)
#ifdef WALLPATCHES
   && GetPatchNum()==lastpatchnum
#endif     
     )
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx = pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }

  if(lastside!=-1) ScalePost();

  lastside=1;
  lastintercept=xtile;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;
  
#ifdef ANIMATEDWALLS
  switch(MAPSPOT(xtile,ytile,1)) 
  { 
    case ANIMTILE: 
      tilehit+=(frameon>>6)%3; 
      break; 
    case ANIMTILE+1: 
      tilehit+=(frameon>>5)%3; 
      break; 
    case ANIMTILE+2: 
      tilehit+=(frameon>>4)%3; 
      break; 
    case ANIMTILE+3: 
      tilehit+=(frameon>>3)%3; 
      break; 
    case ANIMTILE+4: 
      tilehit+=(frameon>>2)%3; 
      break; 
    default: break;
  } 
#endif

  if (tilehit > MAXWALLTILES) tilehit = tilehit - MAXWALLTILES;

  wallpic = vertwall[tilehit];

  if (CheckVertTile (xtile,ytile))
  {                                                               // check for adjacent doors
    if (CheckHighBits (tilemap[xtile-xtilestep][ytile]))
    {
      lasttilehit = 0;
#ifndef DIFFDOORSIDES
      wallpic = DOORWALL+3;
#else
      wallpic = DOORWALL + door_sides[doorobjlist[tilemap[xtile-xtilestep][ytile] & 0x7f].lock] + 1;
#endif
    }
  }
#ifdef WALLPATCHES
//  if (wallpic < DOORWALL && lastpatchnum) 
  if (wallpic < DOORWALL) // && lastpatchnum) 
  {
    if (ApplyPatch (wallpic, 1))
      postsource = patchedwall+texture;
    else
      postsource = PM_GetTexture(wallpic) + texture;
  }
  else
#endif
      postsource = PM_GetTexture(wallpic) + texture;
}

void HitHorizWall (void)
{
  int wallpic;
  int texture;
    
  xtile = (short) (xintercept >> TILESHIFT);
  texture = ((xintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

  if (ytilestep == -1)
    yintercept += TILEGLOBAL;
  else if(MAPSPOT(xtile,ytile,1) != TILEMIRROR)
    texture = TEXTUREMASK-texture;

  if (lastside==0 && lastintercept==ytile && 
      lasttilehit==tilehit && !CheckHorizTile(xtile,ytile)
#ifdef WALLPATCHES
   && GetPatchNum()==lastpatchnum
#endif     
     )
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx=pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }

  if(lastside!=-1) ScalePost();

  lastside=0;
  lastintercept=ytile;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;

#ifdef ANIMATEDWALLS
  switch(MAPSPOT(xtile,ytile,1)) 
  { 
    case ANIMTILE: 
      tilehit+=(frameon>>6)%3; 
      break; 
    case ANIMTILE+1: 
      tilehit+=(frameon>>5)%3; 
      break; 
    case ANIMTILE+2: 
      tilehit+=(frameon>>4)%3; 
      break; 
    case ANIMTILE+3: 
      tilehit+=(frameon>>3)%3; 
      break; 
    case ANIMTILE+4: 
      tilehit+=(frameon>>2)%3; 
      break; 
    default: break;
  } 
#endif

  if (tilehit > MAXWALLTILES) tilehit = tilehit - MAXWALLTILES;

  wallpic = horizwall[tilehit];

  if (CheckHorizTile(xtile,ytile))
  {                                                // check for adjacent doors
    if (CheckHighBits(tilemap[xtile][ytile-ytilestep]))
    {
      lasttilehit = 0;
#ifndef DIFFDOORSIDES
      wallpic = DOORWALL+2;
#else
      wallpic = DOORWALL + door_sides[doorobjlist[tilemap[xtile][ytile-ytilestep] & 0x7f].lock];
#endif
    }
  }
#ifdef WALLPATCHES
//  if (wallpic < DOORWALL && lastpatchnum) 
  if (wallpic < DOORWALL) // && lastpatchnum) 
  {
    if (ApplyPatch (wallpic, 0))
      postsource = patchedwall+texture;
    else
      postsource = PM_GetTexture(wallpic) + texture;
  }
  else
#endif
    postsource = PM_GetTexture(wallpic) + texture;
}

/**************************************************************************
**                         Drawing the Doors                             **
**************************************************************************/ 

#ifdef BLAKESTYLEDOOR
boolean   leftside; 
void SetLeftDoorSide (void)  { leftside=true;  } 
void SetRightDoorSide (void) { leftside=false; } 
#endif

void HitVertDoor (void)
{
    int doornum;
    int texture;
    
    ytile = (short) (yintercept >> TILESHIFT);
    doornum = tilehit&0x7f;

#ifdef BLAKESTYLEDOOR
    if (door_type[doorobjlist[doornum].lock]) 
    { 
      if (leftside)
         texture = ((yintercept+0x7fff-doorobjlist[doornum].lposition) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
      else 
         texture = ((yintercept+0x8000+doorobjlist[doornum].lposition) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
    } 
    else 
      texture = ((yintercept-doorobjlist[doornum].position) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
#else 
    texture = ((yintercept-doorobjlist[doornum].position)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK; 
#endif 

    if(lasttilehit==tilehit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    postsource = PM_GetTexture(DOORWALL + door_tex[doorobjlist[doornum].lock]+1) + texture;
}

void HitHorizDoor (void)
{
    int doornum;
    int texture;
    
    xtile = (short) (xintercept >> TILESHIFT);
    doornum = tilehit&0x7f;

#ifdef BLAKESTYLEDOOR
    if (door_type[doorobjlist[doornum].lock]) 
    { 
      if (leftside)
         texture = ( (xintercept+0x7fff-doorobjlist[doornum].lposition) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
      else 
         texture = ( (xintercept+0x8000+doorobjlist[doornum].lposition) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
    } 
    else 
      texture = ( (xintercept-doorobjlist[doornum].position) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK; 
#else 
    texture = ((xintercept-doorobjlist[doornum].position)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK; 
#endif 
    if(lasttilehit==tilehit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    postsource = PM_GetTexture(DOORWALL + door_tex[doorobjlist[doornum].lock]) + texture;
}


/**************************************************************************
**                        Drawing the SPRITES                            **
**************************************************************************/ 

void DrawScaleds (void)  // Draw Sprites in Players Vision (line of sight)
{
    int       i,least,numvisable,height;
    byte      *tilespot,*visspot;
    unsigned  spotloc;
    statobj_t *statptr;
    objtype   *obj;
    int       spriteadd=0;

    visptr = &vislist[0];

// Animated Objects code
    frameon+=tics;
    frameon%=2520;
    /* remember 70 tics/second, so this will support total cycle time of 2 seconds tops.  Just increase the 140 if you want*/
// End Animated Objects code

//
// place static objects
//
    for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
    {
        if ((visptr->shapenum = statptr->shapenum) == -1) continue; // object has been deleted

        if (!*statptr->visspot) continue;    // Object is not visable

// Animated objects
       switch (visptr->shapenum)
       {
        case TABL_ANIM0:
                           visptr->shapenum+=int((frameon/6)%4);  // Speed of 6, with 4 frames
                           break;
        case SPR_PORTABLE0:
                           visptr->shapenum+=int((frameon/8)%4);  // Speed of 8, with 4 frames
                           break;
        case SPR_DETECT:
                           visptr->shapenum+=int((frameon/12)%2);  // Speed of 12, with 2 frames
                           break;
       }
// End animated objects, now use crate bomb detector
                           if(visptr->shapenum == BOMB_CRATE &&  gamestate.bombdetector) visptr->shapenum+=53;

        if (TransformTile (statptr->tilex,statptr->tiley,&visptr->viewx,&visptr->viewheight) && 
            statptr->flags & FL_BONUS && (statptr->tilex==player->tilex && statptr->tiley==player->tiley))
        {
            GetBonus (statptr);
            if(statptr->shapenum == -1) continue; // object has been taken
        }

        if (!visptr->viewheight) continue; // to close to the object

#ifdef KEYSFLOAT  // BrotherTank's Floating Keys based on Z-Axis Floating
		    switch (visptr->shapenum)
		    {
			    case SPR_STAT_20:
			    case SPR_STAT_21:	
				    if (statptr->position >= 30) statptr->direct = 1; 
				    if (statptr->position <= 1) statptr->direct = 0;
            statptr->position += (!statptr->direct?1:-1);
				    if (statptr->position)
				    {
					    visptr->viewheight += statptr->position; 
				    }
				    break;
		}
#endif

        visptr->transactor=NULL;
#ifdef USE_DIR3DSPR
        if(statptr->flags & FL_DIR_MASK)
            visptr->transsprite=statptr;
        else
            visptr->transsprite=NULL;
#endif
// ************************* Animated Sprite Variations ***********************
        

        // Other Animations Styles
#ifdef FLOATANIMATION // Adam's Floating Keys
        switch (visptr->shapenum)
        {
          case SPR_STAT_20: // Key 1
          case SPR_STAT_21: // Key 2
          // Add more Sprite Names as Above below this line - It's a 4 Frame Animation
            visptr->shapenum += abs(3 - (frameon>>2)%6); 
            break;
          default: break;
        }
#endif

// *********************** Animated Sprite Variations Ends *********************

        if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
        {
            visptr->flags = (short) statptr->flags;
            visptr++;
        }
    }

// place active objects
    for (obj = player->next;obj;obj=obj->next)
    {
        if ((visptr->shapenum = obj->state->shapenum)==0) continue; // no shape

      spriteadd = 0;
      switch (obj->obclass)   
      {
#ifndef DARREN
      case darrenobj:
#endif
//      case dog2obj:
//      case ss2obj:
//      case mutant2obj:
      case purplebloodobj:
        {
         spriteadd = 12;
         break;
         }
      }

        visptr->shapenum+=spriteadd;

        spotloc = (obj->tilex<<mapshift)+obj->tiley;   // optimize: keep in struct?
        visspot = &spotvis[0][0]+spotloc;
        tilespot = &tilemap[0][0]+spotloc;

        // could be in any of the nine surrounding tiles

        if (*visspot || 
           (*(visspot-1) && !*(tilespot-1) )  || (*(visspot+1) && !*(tilespot+1))   ||
           (*(visspot-65) && !*(tilespot-65)) || (*(visspot-64) && !*(tilespot-64)) ||
           (*(visspot-63) && !*(tilespot-63)) || (*(visspot+65) && !*(tilespot+65)) ||
           (*(visspot+64) && !*(tilespot+64)) || (*(visspot+63) && !*(tilespot+63)))
        {
            obj->active = ac_yes;
            TransformActor (obj);
            if (!obj->viewheight) continue;     // too close or far away

            visptr->viewx = obj->viewx;
            visptr->viewheight = obj->viewheight;
            if (visptr->shapenum == -1)
                visptr->shapenum = obj->temp1;  // special shape

            if (obj->state->rotate)
                visptr->shapenum += CalcRotate (obj);

            if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
            {
                visptr->flags = (short) obj->flags;

              visptr->transsprite=NULL;
#ifdef DIRECTIONALACTORS
              visptr->transactor=NULL;

    switch (obj->speed)
           { case -1:
                  visptr->flags = FL_DIR_VERT_MID; visptr->transactor=obj; break;
             case -2:
                  visptr->flags = FL_DIR_HORIZ_MID; visptr->transactor=obj; break; }
#endif
                visptr++;
            }
            obj->flags |= FL_VISABLE;
        }
        else
            obj->flags &= ~FL_VISABLE;
    }
// draw from back to front
    numvisable = (int) (visptr-&vislist[0]);
    if (!numvisable) return;  // no visable objects

    for (i = 0; i<numvisable; i++)
    {
        least = 32000;
        for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
        {
            height = visstep->viewheight;
            if (height < least)
            {
                least = height;
                farthest = visstep;
            }
        }
        // draw farthest
#ifdef USE_DIR3DSPR
    if(farthest->flags & FL_DIR_POS_MASK) {
        if(farthest->transactor != NULL) Scale3DShapeActor(vbuf, vbufPitch, farthest->transactor, farthest->flags);
        else if(farthest->transsprite) Scale3DShape(vbuf, vbufPitch, farthest->transsprite); }
        else
#endif
            ScaleShape(farthest->viewx, farthest->shapenum, farthest->viewheight, farthest->flags);

        farthest->viewheight = 32000;
    }
}

/**************************************************************************
**                  Draw the Players Hands & Weapons                     **
**************************************************************************/ 
void DrawPlayerWeapon (void)
{
    int shapenum;

#ifndef SPEAR
    if (gamestate.victoryflag)
    {
#ifndef APOGEE_1_0
        if (player->state == &s_deathcam && (GetTimeCount()&32) )
            SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
#endif
        return;
    }
#endif
    if (gamestate.weapon != -1)
    {
        shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
        SimpleScaleShape(viewwidth/2,shapenum,viewheight+1);
    }
    if (demorecord || demoplayback)
        SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
}

// Chris's raycaster bugfix
void RefreshCorner (void)
{
    int xpart = (xintercept&(TILEGLOBAL-1)) >> 10;
    int ypart = (yintercept&(TILEGLOBAL-1)) >> 10;

    if ((xtilestep == -1 && !xpart) || (xtilestep == 1 && xpart == 63))
        if ((ytilestep == -1 && !ypart) || (ytilestep == 1 && ypart == 63))
            tilehit = lasttilehit;
}

/**************************************************************************
**          ASMREFRESH - The Heart of the Drawing Routines               **
**************************************************************************/ 
void AsmRefresh()
{
  int32_t xstep,ystep;
  longword xpartial,ypartial;
  boolean playerInPushwallBackTile = tilemap[focaltx][focalty] == 255;

  for(pixx=0;pixx<viewwidth;pixx++)
  {
    short angl=midangle+pixelangle[pixx];
    if(angl<0) angl+=FINEANGLES;
    if(angl>=3600) angl-=FINEANGLES;
    if(angl<900)
    {
      xtilestep=1;
      ytilestep=-1;
      xstep=finetangent[900-1-angl];
      ystep=-finetangent[angl];
      xpartial=xpartialup;
      ypartial=ypartialdown;
    }
    else if(angl<1800)
    {
      xtilestep=-1;
      ytilestep=-1;
      xstep=-finetangent[angl-900];
      ystep=-finetangent[1800-1-angl];
      xpartial=xpartialdown;
      ypartial=ypartialdown;
    }
    else if(angl<2700)
    {
      xtilestep=-1;
      ytilestep=1;
      xstep=-finetangent[2700-1-angl];
      ystep=finetangent[angl-1800];
      xpartial=xpartialdown;
      ypartial=ypartialup;
    }
    else if(angl<3600)
    {
      xtilestep=1;
      ytilestep=1;
      xstep=finetangent[angl-2700];
      ystep=finetangent[3600-1-angl];
      xpartial=xpartialup;
      ypartial=ypartialup;
    }
    yintercept=FixedMul(ystep,xpartial)+viewy;
    xtile=focaltx+xtilestep;
    xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
    xintercept=FixedMul(xstep,ypartial)+viewx;
    ytile=focalty+ytilestep;
    yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
    texdelta=0;
    
    // Special treatment when player is in back tile of pushwall
    if(playerInPushwallBackTile)
    {
      if(pwall.dir == di_east && xtilestep ==  1 || 
         pwall.dir == di_west && xtilestep == -1)
      {
        int32_t yintbuf = yintercept - ((ystep * (64 - pwall.pos)) >> 6);
        if((yintbuf >> 16) == focalty)   // ray hits pushwall back?
        {
          if(pwall.dir == di_east)
            xintercept = (focaltx << TILESHIFT) + (pwall.pos << 10);
          else
            xintercept = (focaltx << TILESHIFT) - TILEGLOBAL + ((64 - pwall.pos) << 10);

          yintercept = yintbuf;
          tilehit = pwall.tile;
          HitVertWall();
          continue;
        }
      }
      else if(pwall.dir == di_south && ytilestep ==  1 ||
              pwall.dir == di_north && ytilestep == -1)
      {
        int32_t xintbuf = xintercept - ((xstep * (64 - pwall.pos)) >> 6);
        if((xintbuf >> 16) == focaltx)   // ray hits pushwall back?
        {
          xintercept = xintbuf;
          if(pwall.dir == di_south)
            yintercept = (focalty << TILESHIFT) + (pwall.pos << 10);
          else
            yintercept = (focalty << TILESHIFT) - TILEGLOBAL + ((64 - pwall.pos) << 10);

          tilehit = pwall.tile;
          HitHorizWall();
          continue;
        }
      }
    }
// Vertical Tile Detection
    do
    {
      if(ytilestep==-1 && (yintercept>>16)<=ytile) goto horizentry;
      if(ytilestep==1 && (yintercept>>16)>=ytile) goto horizentry;
vertentry:
      if((uint32_t)yintercept>mapheight*65536-1 || (word)xtile>=mapwidth)
      {
        if(xtile<0) xintercept=0;
        if(xtile>=mapwidth) xintercept=mapwidth<<TILESHIFT;
        if(yintercept<0) yintercept=0;
        if(yintercept>=(mapheight<<TILESHIFT)) yintercept=mapheight<<TILESHIFT;
        yspot=0xffff;
        tilehit=0;
        HitHorizWall(); // Horizontal Border of Map
        break;
      }
      if(xspot>=maparea) break;
      tilehit=((byte *)tilemap)[xspot];
      if (tilemap[xtile][ytile]) RefreshCorner();
      if(tilehit)
      {  
        if(tilehit==255)  // Found a Pushwall
        { 
          if(pwall.dir==di_west || pwall.dir==di_east)
          {
            int32_t yintbuf;
            int pwallposnorm;
            int pwallposinv;
            if(pwall.dir==di_west)
            {
              pwallposnorm = 64-pwall.pos;
              pwallposinv = pwall.pos;
            }
            else
            {
              pwallposnorm = pwall.pos;
              pwallposinv = 64-pwall.pos;
            }
            if(pwall.dir == di_east && xtile==pwall.x && ((uint32_t)yintercept>>16)==pwall.y ||
               pwall.dir == di_west && !(xtile==pwall.x && ((uint32_t)yintercept>>16)==pwall.y))
            {
              yintbuf=yintercept+((ystep*pwallposnorm)>>6);
              if((yintbuf>>16)!=(yintercept>>16)) goto passvert;

              xintercept=(xtile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
              yintercept=yintbuf;
              tilehit=pwall.tile;
              HitVertWall();
            }
            else
            {
              yintbuf=yintercept+((ystep*pwallposinv)>>6);
              if((yintbuf>>16)!=(yintercept>>16)) goto passvert;

              xintercept=(xtile<<TILESHIFT)-(pwallposinv<<10);
              yintercept=yintbuf;
              tilehit=pwall.tile;
              HitVertWall();
            }
          }
          else
          {
            int pwallposi = pwall.pos;
            if(pwall.dir==di_north) pwallposi = 64-pwall.pos;
            if(pwall.dir==di_south && (word)yintercept<(pwallposi<<10) || 
               pwall.dir==di_north && (word)yintercept>(pwallposi<<10))
            {
              if(((uint32_t)yintercept>>16)==pwall.y && xtile==pwall.x)
              {
                if(pwall.dir==di_south && 
                  (int32_t)((word)yintercept)+ystep<(pwallposi<<10) || 
                   pwall.dir==di_north && 
                  (int32_t)((word)yintercept)+ystep>(pwallposi<<10)) goto passvert;

                if(pwall.dir==di_south)
                  yintercept=(yintercept&0xffff0000)+(pwallposi<<10);
                else
                  yintercept=(yintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);

                xintercept=xintercept-((xstep*(64-pwall.pos))>>6);
                tilehit=pwall.tile;
                HitHorizWall();
              }
              else
              {
                texdelta = -(pwallposi<<10);
                xintercept=xtile<<TILESHIFT;
                tilehit=pwall.tile;
                HitVertWall();
              }
            }
            else
            {
              if(((uint32_t)yintercept>>16)==pwall.y && xtile==pwall.x)
              {
                texdelta = -(pwallposi<<10);
                xintercept=xtile<<TILESHIFT;
                tilehit=pwall.tile;
                HitVertWall();
              }
              else
              {
                if(pwall.dir==di_south && 
                  (int32_t)((word)yintercept)+ystep>(pwallposi<<10) || 
                   pwall.dir==di_north && 
                  (int32_t)((word)yintercept)+ystep<(pwallposi<<10)) goto passvert;

                if(pwall.dir==di_south)
                  yintercept=(yintercept&0xffff0000)-((64-pwall.pos)<<10);
                else
                  yintercept=(yintercept&0xffff0000)+((64-pwall.pos)<<10);

                xintercept=xintercept-((xstep*pwall.pos)>>6);
                tilehit=pwall.tile;
                HitHorizWall();
              }
            }
          }
        }
        else if(tilehit&0x80) // Found a Door
        {
          int32_t yintbuf=yintercept+(ystep>>1);
            
          if((yintbuf>>16)!=(yintercept>>16)) goto passvert;
#ifdef BLAKESTYLEDOOR 
          SetLeftDoorSide(); 
          if((word)yintbuf<=doorobjlist[tilehit&0x7f].lposition) goto drawvertdoor; 
          SetRightDoorSide(); 
          if((word)yintbuf<doorobjlist[tilehit&0x7f].position) goto passvert;
drawvertdoor:
#else 
          if((word)yintbuf<doorobjlist[tilehit&0x7f].position) goto passvert;
#endif 
          yintercept=yintbuf;
          xintercept=(xtile<<TILESHIFT)|0x8000;
          HitVertDoor();
        }
        else  // Normal Wall Tile then
        {
          xintercept=xtile<<TILESHIFT;
          HitVertWall();
        }
        break;
      }
passvert:
      *((byte *)spotvis+xspot)=1;
      xtile+=xtilestep;
      yintercept+=ystep;
      xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
    }
    while(1);
    continue;
// Horizontal Tile Detection
    do
    {
      if(xtilestep==-1 && (xintercept>>16)<=xtile) goto vertentry;
      if(xtilestep==1 && (xintercept>>16)>=xtile) goto vertentry;
horizentry:
      if((uint32_t)xintercept>mapwidth*65536-1 || (word)ytile>=mapheight)
      {
        if(ytile<0) yintercept=0;
        if(ytile>=mapheight) yintercept=mapheight<<TILESHIFT;
        if(xintercept<0) xintercept=0;
        if(xintercept>=(mapwidth<<TILESHIFT)) xintercept=mapwidth<<TILESHIFT;
        xspot=0xffff;
        tilehit=0;
        HitVertWall();  // Vertical Border of Map
        break;
      }
      if(yspot>=maparea) break;
      tilehit=((byte *)tilemap)[yspot];
      if (tilemap[xtile][ytile]) RefreshCorner();
      if(tilehit)
      {
        if(tilehit==255)
        {
          if(pwall.dir==di_north || pwall.dir==di_south)
          {
            int32_t xintbuf;
            int pwallposnorm;
            int pwallposinv;
            if(pwall.dir==di_north)
            {
              pwallposnorm = 64-pwall.pos;
              pwallposinv = pwall.pos;
            }
            else
            {
              pwallposnorm = pwall.pos;
              pwallposinv = 64-pwall.pos;
            }
            if(pwall.dir == di_south && ytile==pwall.y && 
              ((uint32_t)xintercept>>16)==pwall.x || 
               pwall.dir == di_north && !(ytile==pwall.y && 
              ((uint32_t)xintercept>>16)==pwall.x))
            {
              xintbuf=xintercept+((xstep*pwallposnorm)>>6);
              if((xintbuf>>16)!=(xintercept>>16)) goto passhoriz;
  
              yintercept=(ytile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
              xintercept=xintbuf;
              tilehit=pwall.tile;
              HitHorizWall();
            }
            else
            {
              xintbuf=xintercept+((xstep*pwallposinv)>>6);
              if((xintbuf>>16)!=(xintercept>>16)) goto passhoriz;
  
              yintercept=(ytile<<TILESHIFT)-(pwallposinv<<10);
              xintercept=xintbuf;
              tilehit=pwall.tile;
              HitHorizWall();
            }
          }
          else
          {
            int pwallposi = pwall.pos;
            if(pwall.dir==di_west) pwallposi = 64-pwall.pos;
            if(pwall.dir==di_east && (word)xintercept<(pwallposi<<10) || 
               pwall.dir==di_west && (word)xintercept>(pwallposi<<10))
            {
              if(((uint32_t)xintercept>>16)==pwall.x && ytile==pwall.y)
              {
                if(pwall.dir==di_east && (int32_t)((word)xintercept)+xstep<(pwallposi<<10) || 
                   pwall.dir==di_west && (int32_t)((word)xintercept)+xstep>(pwallposi<<10)) goto passhoriz;

                if(pwall.dir==di_east)
                  xintercept=(xintercept&0xffff0000)+(pwallposi<<10);
                else
                  xintercept=(xintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);

                yintercept=yintercept-((ystep*(64-pwall.pos))>>6);
                tilehit=pwall.tile;
                HitVertWall();
              }
              else
              {
                texdelta = -(pwallposi<<10);
                yintercept=ytile<<TILESHIFT;
                tilehit=pwall.tile;
                HitHorizWall();
              }
            }
            else
            {
              if(((uint32_t)xintercept>>16)==pwall.x && ytile==pwall.y)
              {
                texdelta = -(pwallposi<<10);
                yintercept=ytile<<TILESHIFT;
                tilehit=pwall.tile;
                HitHorizWall();
              }
              else
              {
                if(pwall.dir==di_east && (int32_t)((word)xintercept)+xstep>(pwallposi<<10) || 
                   pwall.dir==di_west && (int32_t)((word)xintercept)+xstep<(pwallposi<<10)) goto passhoriz;
  
                if(pwall.dir==di_east)
                  xintercept=(xintercept&0xffff0000)-((64-pwall.pos)<<10);
                else
                  xintercept=(xintercept&0xffff0000)+((64-pwall.pos)<<10);

                yintercept=yintercept-((ystep*pwall.pos)>>6);
                tilehit=pwall.tile;
                HitVertWall();
              }
            }
          }
        }
        else if(tilehit&0x80)
        {
          int32_t xintbuf=xintercept+(xstep>>1);
          if((xintbuf>>16)!=(xintercept>>16)) goto passhoriz;
          
#ifdef BLAKESTYLEDOOR 
          SetLeftDoorSide(); 
          if((word)xintbuf<=doorobjlist[tilehit&0x7f].lposition) goto drawhorizdoor; 
          SetRightDoorSide(); 
          if((word)xintbuf<doorobjlist[tilehit&0x7f].position) goto passhoriz;
drawhorizdoor: 
#else 
          if((word)xintbuf<doorobjlist[tilehit&0x7f].position) goto passhoriz;
#endif 
          xintercept=xintbuf;
          yintercept=(ytile<<TILESHIFT)+0x8000;
          HitHorizDoor();
        }
        else
        {
          yintercept=ytile<<TILESHIFT;
          HitHorizWall();
        }
        break;
      }
passhoriz:
      *((byte *)spotvis+yspot)=1;
      ytile+=ytilestep;
      xintercept+=xstep;
      yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
    }
    while(1);
  }
}

#ifdef CHRISAUTOMAP
#ifdef CHRISAUTOMAP2
boolean BeenThere [MAPSIZE][MAPSIZE];
#endif 
/**************************************************************************
**                        Draw Map over Weapon Pic                       **
**************************************************************************/ 
void DrawLevelMap (void) 
{ 
   int x=player->tilex,y=player->tiley; 
   int offx,offy,midx=249,midy=165,lenx=61,leny=31; 
   offx = x*4/mapwidth; offy = y*34/mapheight; 

   for (x=0;x<lenx;x++) 
      for (y=0;y<leny;y++) 
      { 
#ifdef CHRISAUTOMAP2
         if (spotvis[offx+x][offy+y] || BeenThere [offx+x][offy+y]) 
         {
           if (BeenThere [offx+x][offy+y] != 1) BeenThere [offx+x][offy+y] = 1;
#else        
         if (tilemap[offx+x][offy+y] == 0) 
         {
#endif
            VWB_Plot(midx+x,midy+y,0); 
            if (actorat[offx+x][offy+y] != 0) 
            { 
               if ((unsigned)actorat[offx+x][offy+y] == 1) 
                  VWB_Plot(midx+x,midy+y,158); 
               else 
                  VWB_Plot(midx+x,midy+y,150); 
            } 
         } 
         else 
            VWB_Plot(midx+x,midy+y,154); 
      } 

   VWB_Plot(player->tilex+midx-offx,player->tiley+midy-offy,140); 
   VW_UpdateScreen (); 
} 
#else
  #ifdef CHRISAUTOMAPTEK
boolean WalkedFloor[NUMAREAS]; 

void DrawWeapon (void) 
{ 
   VL_Bar(249,165,61,31,0x97); 
    
   WalkedFloor[player->areanumber] = true; 
    
   if (!Keyboard[sc_O]) 
      StatusDrawPic (32,8,KNIFEPIC+gamestate.weapon); 
   else if (!Keyboard[sc_Tab]) 
   { 
      int x=player->tilex,y=player->tiley; 
      int offx,offy,midx=249,midy=165,lenx=61,leny=31; 
      offx = x*4/64; offy = y*34/64; 
      for (x=0;x<lenx;x++) 
         for (y=0;y<leny;y++) 
         { 
            if (tilemap[offx+x][offy+y] == 0) 
            { 
               if (WalkedFloor[MAPSPOT(offx+x,offy+y,0)-AREATILE]) 
                  VL_Plot(midx+x,midy+y,0x00); 
               else 
               { 
                  VL_Plot(midx+x,midy+y,0x97); 
                  continue; 
               } 
               if (actorat[offx+x][offy+y] != NULL) 
               { 
                  if ((unsigned)actorat[offx+x][offy+y] == 1) 
                     VL_Plot(midx+x,midy+y,0x80); 
                  else 
                     VL_Plot(midx+x,midy+y,0x77); 
               } 
            } 
            else 
               VL_Plot(midx+x,midy+y,0x97); 
         } 
      VL_Plot(player->tilex+midx-offx,player->tiley+midy-offy,0x0F); 
   } 
} 
  #endif
#endif

/**************************************************************************
**                              WallRefresh                              **
**************************************************************************/ 
void WallRefresh (void)
{
  xpartialdown = viewx&(TILEGLOBAL-1);
  xpartialup = TILEGLOBAL-xpartialdown;
  ypartialdown = viewy&(TILEGLOBAL-1);
  ypartialup = TILEGLOBAL-ypartialdown;
  min_wallheight = viewheight;
  lastside = -1;                  // the first pixel is on a new wall
  AsmRefresh ();
  ScalePost ();                   // no more optimization on last post
}

/**************************************************************************
**         CalcViewVariables - From the Players Perspective              **
**************************************************************************/ 
void CalcViewVariables()
{
#ifdef TEKZOOMSHOCK
  viewangle = GetShockedAngle(); 
#else
  viewangle = player->angle;
#endif
  midangle = viewangle*(FINEANGLES/ANGLES);
  viewsin = sintable[viewangle];
  viewcos = costable[viewangle];
#ifdef TEKZOOMSHOCK
  if ( focallength < FOCALTHRESHOLD ) 
  { 
    viewx = player->x - FixedMul( focallength, viewcos ); 
    viewy = player->y + FixedMul( focallength, viewsin ); 
  } 
  else 
  { 
    viewx = player->x - FixedMul( FOCALTHRESHOLD, viewcos ); 
    viewy = player->y + FixedMul( FOCALTHRESHOLD, viewsin ); 
  } 
#else
  viewx = player->x - FixedMul(focallength,viewcos);
  viewy = player->y + FixedMul(focallength,viewsin);
#endif
  focaltx = (short)(viewx>>TILESHIFT);
  focalty = (short)(viewy>>TILESHIFT);
  viewtx = (short)(player->x >> TILESHIFT);
  viewty = (short)(player->y >> TILESHIFT);
}

/**************************************************************************
**         ThreeDRefresh - Controlling the Drawing Process               **
**************************************************************************/ 
void ThreeDRefresh (void)
{

  memset(spotvis,0,maparea);                      // clear out the traced array
  spotvis[player->tilex][player->tiley] = 1;      // Detect all sprites over player fix

  vbuf = VL_LockSurface(screenBuffer);
  vbuf+=screenofs;
  vbufPitch = bufferPitch;

  CalcViewVariables();

// follow the walls from there to the right, drawing as we go

  VGAClearScreen ();

#if defined(USE_FEATUREFLAGS) && defined(USE_STARSKY)
  if(GetFeatureFlags() & FF_STARSKY) DrawStarSky(vbuf, vbufPitch);
#endif

  WallRefresh ();

#if defined(USE_FEATUREFLAGS) && defined(USE_PARALLAX)
  if(GetFeatureFlags() & FF_PARALLAXSKY) DrawParallax(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_CLOUDSKY)
  if(GetFeatureFlags() & FF_CLOUDSKY) DrawClouds(vbuf, vbufPitch, min_wallheight);
#endif

#if defined(USE_FEATUREFLAGS) && defined(USE_RAIN)
  if(GetFeatureFlags() & FF_RAIN) DrawRain(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_SNOW)
  if(GetFeatureFlags() & FF_SNOW) DrawSnow(vbuf, vbufPitch);
#endif

#ifdef USE_FLOORCEILINGTEX
if(!ffDataBottomRight)
  DrawFloorAndCeiling(vbuf, vbufPitch, min_wallheight);
#endif

// draw all the scaled images
  DrawScaleds();                  // draw scaled stuff

  DrawPlayerWeapon ();    // draw player's hands

if (messagetime > 0)
   DrawMessage ();

  if(Keyboard[sc_Tab] && viewsize == 21 && gamestate.weapon != -1)
    ShowActStatus();

  VL_UnlockSurface(screenBuffer);
  vbuf = NULL;

// show screen and time last cycle
  if (fizzlein)
  {
    FizzleFade(screenBuffer, screen, 0, 0, screenWidth, screenHeight, 20, false);
    fizzlein = false;

    lasttimecount = GetTimeCount();          // don't make a big tic count
  }
  else
  {
#ifndef REMDEBUG
    if (fpscounter)
    {
      fontnumber = 0;
      SETFONTCOLOR(7,127);
      PrintX=4; PrintY=1;
      VWB_Bar(0,0,50,10,bordercol);
      US_PrintSigned(fps);
      US_Print(" fps");
    }
#endif
    SDL_BlitSurface(screenBuffer, NULL, screen, NULL);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
  }

#ifndef REMDEBUG
  if (fpscounter)
  {
    fps_frames++;
    fps_time+=tics;

    if(fps_time>35)
    {
      fps_time-=35;
      fps=fps_frames<<1;
      fps_frames=0;
    }
  }
#endif
}

/*
========================
=
= GetMessage
=
= gets ingame messages
=
========================
*/

void GetMessage (char *lastmessage)
{
   if(viewwidth == screenWidth) {
   messagetime = 150; // time for message to display

   strcpy(gamestate.message, lastmessage);
      strcpy(gamestate.message, lastmessage);
//      TextXVar = ((screenWidth-viewwidth)/4)+2;
//      TextYVar = ((TextXVar-2)/2)+1;
//      if(viewwidth == screenWidth) { TextXVar = 2; TextYVar = 2; }
      TextXVar = 2; TextYVar = 2; }
}   

/*
========================
=
= DrawMessage
=
= displays ingame messages
=
========================
*/

void DrawMessage (void)
{
   messagetime-=tics;

   fontnumber = 0;
   SETFONTCOLOR(0xf,0x0); // set the color
   PrintX=TextXVar; PrintY=TextYVar; // position the message
   US_Print(gamestate.message); // print message
   DrawPlayBorderSides ();
}
